/* $Id: crires_load.c,v 1.62 2012-09-19 14:10:27 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-19 14:10:27 $
 * $Revision: 1.62 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_wlxcorr.h"

#include "crires_load.h"
#include "crires_dfs.h"
#include "crires_pfits.h"

/*-----------------------------------------------------------------------------
   								Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_table_check_column(const cpl_table *, const char *) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_load    Loading related functions
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the total number of NDIT * DIT
  @param    fname       the input file name
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
double crires_get_totndit(const char * fname) 
{
    cpl_propertylist    *   plist ;
    int                     ndit ;
    double                  dit ;
    
    /* Test entries */
    if (fname == NULL) return -1 ;

    if ((plist=cpl_propertylist_load(fname, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot get header from input file") ;
        return -1 ;
    }
    ndit = crires_pfits_get_ndit(plist) ;
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "No NDIT/DIT information in the header");
        return -1 ;
    }
    return ndit * dit ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the chip of CRIRES images from a frameset in an image list
  @param    fset        the input set of frames
  @param    period
  @param    chip        the chip to load (1 to CRIRES_NB_DETECTORS)
  @param    ptype       the pixel type
  @return   the CRIRES_NB_DETECTORS images image list or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_load_frameset(
        const cpl_frameset  *   fset, 
        crires_illum_period     period,
        int                     chip,
        cpl_type                ptype) 
{
    cpl_imagelist       *   out ;
    const cpl_frame     *   frame ;
    const char          *   fname ;
    cpl_image           *   ima ;
    int                     i ;
    
    /* Test entries */
    if (fset == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ; 

    /* Create the output imagelist */
    out = cpl_imagelist_new() ;
    
    /* Loop on the frames */
    for (i=0 ; i<cpl_frameset_get_size(fset) ; i++) {
        /* Load the ith frame */
        frame = cpl_frameset_get_position_const(fset, i) ;
        fname = cpl_frame_get_filename(frame) ;
        if ((ima=crires_load_chip(fname, period, chip, ptype)) == NULL) {
            cpl_msg_error(__func__, "Cannot load %dth frame (chip %d)", 
                    i+1, chip) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
        cpl_imagelist_set(out, ima, i) ;
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load a quadrant CRIRES image in an image 
  @param    fname       the input file name
  @param    period
  @param    chip        the chip nb (1 to CRIRES_NB_DETECTORS)
  @param    ptype       the pixel type
  @return   the image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_load_chip(
        const char      *   fname,
        crires_illum_period period,
        int                 chip,
        cpl_type            ptype) 
{
    return crires_load_chip_plane(fname, period, chip, 0, ptype) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load stripes image from a full detector multi extensions file
  @param    fname       the input file name
  @param    chip        the chip nb (2 to CRIRES_NB_DETECTORS-1)
  @param    plane       the plane number
  @param    ptype       the pixel type
  @param    center_y    the y center position of the lower stripe
  @param    offset_y    the y offset between the 2 stripes centers
  @param    size_y      the y size of a stripe
  @return   the image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_load_chip_plane_stripes(
        const char      *   fname,
        int                 chip,
        int                 plane,
        cpl_type            ptype,
        int                 center_y,
        int                 offset_y,
        int                 size_y) 
{
    cpl_image           *   out ;
    cpl_image           *   stripe ;
    int                     y1, y2, y3, y4 ;

    /* Test entries */
    if (fname == NULL) return NULL ;
    if (chip < 2 || chip > CRIRES_NB_DETECTORS-1) return NULL ;
    if (center_y < 0 || offset_y < 0 || size_y < 0) return NULL ;

    /* Initialise */
    y1 = (int)(center_y - (size_y/2.0) + 1) ;
    y2 = y1 + size_y - 1 ;
    y3 = y1 + offset_y ;
    y4 = y3 + size_y - 1 ;

    /* Create the output image */
    out = cpl_image_new(1024, 2*size_y, ptype) ;

    /* Extract the lower stripe */
    if ((stripe = cpl_image_load_window(fname, ptype, plane, chip, 
                    1, y1, 1024, y2)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the chip %d lower part from %s",
                chip, fname) ;
        cpl_image_delete(out) ;
        return NULL ;
    }

    /* Copy the lower part */
    if (cpl_image_copy(out, stripe, 1, 1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the lower part") ;
        cpl_image_delete(out) ;
        cpl_image_delete(stripe) ;
        return NULL ;
    }
    cpl_image_delete(stripe) ;

    /* Extract the upper stripe */
    if ((stripe = cpl_image_load_window(fname, ptype, plane, chip, 
                    1, y3, 1024, y4)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the chip %d upper part from %s",
                chip, fname) ;
        cpl_image_delete(out) ;
        return NULL ;
    }

    /* Copy the upper part */
    if (cpl_image_copy(out, stripe, 1, size_y + 1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the upper part") ;
        cpl_image_delete(out) ;
        cpl_image_delete(stripe) ;
        return NULL ;
    }
    cpl_image_delete(stripe) ;

    /* return  */
    return out ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Load a quadrant CRIRES image from a cube in an image 
  @param    fname       the input file name
  @param    period
  @param    chip        the chip nb (1 to CRIRES_NB_DETECTORS)
  @param    plane       the plane number
  @param    ptype       the pixel type
  @return   the image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_load_chip_plane(
        const char      *   fname,
        crires_illum_period period,
        int                 chip,
        int                 plane,
        cpl_type            ptype) 
{
    cpl_image           *   out ;
    cpl_propertylist    *   plist ;
    int                     uy, ly, ny ;
        
    /* Initialise */
    ly = uy = -1 ;

    /* Test entries */
    if (fname == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    if (period == CRIRES_ILLUM_FULL_DETECTOR) {
        /* Load the full detector */
        if ((out = cpl_image_load(fname, ptype, plane, chip)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the chip %d from %s",
                    chip, fname) ;
            return NULL ;
        }
    } else if ((period == CRIRES_ILLUM_LOWER_HALF) ||
            (period == CRIRES_ILLUM_UPPER_HALF)) {
        /* Load half of the detector - Get the Y size */
        if ((plist=cpl_propertylist_load(fname, chip)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the property list") ;
            return NULL ;
        }
        ny = crires_pfits_get_naxis2(plist) ;
        cpl_propertylist_delete(plist) ;

        if (period == CRIRES_ILLUM_LOWER_HALF) {
            /* Load the lower half of the detector */
            ly = 1 ;
            uy = (ny / 2) ;
        } else if (period == CRIRES_ILLUM_UPPER_HALF) {
            /* Load the upper half of the detector */
            ly = (ny / 2) + 1 ;
            uy = ny ;
        }

        /* Load */
        if ((out = cpl_image_load_window(fname, ptype, plane, chip, 
                        1, ly, 1024, uy)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the chip %d from %s",
                    chip, fname) ;
            return NULL ;
        }
    } else {
        /* Detector partly illuminated */
        if (chip == 1) {
            ly = crires_get_detector_ly1(period) ;
            uy = crires_get_detector_uy1(period) ;
        }
        if (chip == 2) {
            ly = crires_get_detector_ly2(period) ;
            uy = crires_get_detector_uy2(period) ;
        }
        if (chip == 3) {
            ly = crires_get_detector_ly3(period) ;
            uy = crires_get_detector_uy3(period) ;
        }
        if (chip == 4) {
            ly = crires_get_detector_ly4(period) ;
            uy = crires_get_detector_uy4(period) ;
        }
        if ((out = cpl_image_load_window(fname, ptype, plane, chip, 
                        1, ly, 1024, uy)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the chip %d from %s",
                    chip, fname) ;
            return NULL ;
        }
    }

    /* return  */
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load a CRIRES_NB_DETECTORS quadrant CRIRES image in an image list
  @param    fname       the input file name
  @param    period
  @param    ptype       the pixel type
  @return   the CRIRES_NB_DETECTORS images image list or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_load_file(
        const char          *   fname, 
        crires_illum_period     period,
        cpl_type                ptype) 
{
    cpl_imagelist       *   out ;
    cpl_image           *   extracted ;
    int                     i ;
    
    /* Test entries */
    if (fname == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;

    /* Create the output cube */
    out = cpl_imagelist_new() ;

    /* Extract the CRIRES_NB_DETECTORS zones */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if ((extracted = crires_load_chip(fname, period, i+1, ptype)) == NULL) {
            cpl_imagelist_delete(out) ;
            return NULL ;
        }

        if (cpl_imagelist_set(out, extracted, i) != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot add the current image to the list");
            cpl_image_delete(extracted) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Check that the passed file is the proper one and load it
  @param    in          the input file
  @param    extnum      the extension number
  @param    protype     the expected pro.type
  @param    pmin        the first pixel to load (-1 if all)
  @param    pmax        the last pixel to load (-1 if all)
  @param    model_purge flag to purge the wavelength_model = 0.0 values
  @return   the loaded table or NULL if not the proper one.
  
  The function checks the PRO TYPE, and that the expected columns are there
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_load_table_check(
        const char  *   in,
        int             extnum,
        const char  *   protype,
        int             pmin,
        int             pmax,
        int             model_purge)
{
    cpl_table           *   out ;
    cpl_table           *   out_tmp ;
    cpl_propertylist    *   plist ;
    const char          *   sval ;
    int                     nok ;

    /* Check entries */
    if (in == NULL) return NULL ;
    if (protype == NULL) return NULL ;
    if (extnum < 1 || extnum > CRIRES_NB_DETECTORS) return NULL ;

    /* Check that the file contains the proper PRO.TYPE */
    if ((plist=cpl_propertylist_load(in, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from file %s", in);
        return NULL ;
    }
    if ((sval = crires_pfits_get_protype(plist)) == NULL) {
        cpl_msg_error(__func__, "No PRO.TYPE in file %s", in);
        cpl_propertylist_delete(plist) ;
        return NULL ;
    }
    if (strcmp(sval, protype)) {
        cpl_propertylist_delete(plist) ;
        return NULL ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Load the table */
    if ((out = cpl_table_load(in, extnum, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot load %s as a table", in) ;
        return NULL ;
    }

    /* Check that the columns are there */
    nok = 0 ;
    if (!strcmp(protype, CRIRES_PROTYPE_CATALOG)) {
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH) ;
        nok += crires_table_check_column(out, CRIRES_COL_EMISSION) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_PHO_FLUX)) {
        nok += crires_table_check_column(out, CRIRES_COL_STDNAME) ;
        nok += crires_table_check_column(out, CRIRES_COL_RA) ;
        nok += crires_table_check_column(out, CRIRES_COL_DEC) ;
        nok += crires_table_check_column(out, CRIRES_COL_PHOTOFLUX) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_SPEC_PIX)) {
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_BG_NOISE) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_SPEC_WL)) {
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_BG_NOISE) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH_MODEL) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_CONVERS)) {
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_BG_NOISE) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH_MODEL) ;
        nok += crires_table_check_column(out, CRIRES_COL_CONVERSION_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_CONVERSION_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_THROUGHPUT) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_SENSIT)) {
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_INT_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_ERR_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_WN_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_EXTRACT_BG_NOISE) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH) ;
        nok += crires_table_check_column(out, CRIRES_COL_WAVELENGTH_MODEL) ;
        nok += crires_table_check_column(out, CRIRES_COL_CONVERSION_RECT) ;
        nok += crires_table_check_column(out, CRIRES_COL_CONVERSION_OPT) ;
        nok += crires_table_check_column(out, CRIRES_COL_THROUGHPUT) ;
        nok += crires_table_check_column(out, CRIRES_COL_SENSITIVITY) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_WL_POLY)) {
        /* CRIRES_COL_WL_POSY is optional */
        /* nok += crires_table_check_column(out, CRIRES_COL_WL_POSY) ; */
        /* CRIRES_COL_WL_XCORR is optional */
        /* nok += crires_table_check_column(out, CRIRES_COL_WL_XCORR) ; */
        nok += crires_table_check_column(out, CRIRES_COL_WL_COEFFA) ;
        nok += crires_table_check_column(out, CRIRES_COL_WL_COEFFB) ;
        nok += crires_table_check_column(out, CRIRES_COL_WL_COEFFC) ;
        nok += crires_table_check_column(out, CRIRES_COL_WL_COEFFD) ;
        nok += crires_table_check_column(out, CRIRES_COL_WL_COEFFE) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_XCORR)) {
        nok += crires_table_check_column(out, IRPLIB_WLXCORR_COL_WAVELENGTH) ;
        nok += crires_table_check_column(out, IRPLIB_WLXCORR_COL_CAT_INIT) ;
        nok += crires_table_check_column(out, IRPLIB_WLXCORR_COL_CAT_FINAL) ;
        nok += crires_table_check_column(out, IRPLIB_WLXCORR_COL_OBS) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_FWHM)) {
        nok += crires_table_check_column(out, CRIRES_COL_WL_DETPOSX) ;
        nok += crires_table_check_column(out, CRIRES_COL_DETPOSY) ;
        nok += crires_table_check_column(out, CRIRES_COL_FWHM) ;
    } else if (!strcmp(protype, CRIRES_PROTYPE_MOD_CONF)) {
        nok += crires_table_check_column(out, CRIRES_COL_MODEL_CONF_NAME) ;
        nok += crires_table_check_column(out, CRIRES_COL_MODEL_CONF_BEST) ;
        nok += crires_table_check_column(out, CRIRES_COL_MODEL_CONF_LOW) ;
        nok += crires_table_check_column(out, CRIRES_COL_MODEL_CONF_HIGH) ;
        nok += crires_table_check_column(out, CRIRES_COL_MODEL_CONF_FLAG) ;
    } else {
        cpl_msg_error(__func__, "Unsupported PRO.TYPE: %s", protype) ;
        cpl_table_delete(out) ;
        return NULL ;
    }

    /* Check if ok */
    if (nok > 0) {
        cpl_table_delete(out) ;
        return NULL ;
    }

    /* Select only between pmin and pmax */
    if (pmin>0 && pmax>0 && pmax >= pmin && pmax <= cpl_table_get_nrow(out)) {
        out_tmp = cpl_table_extract(out, pmin, pmax-pmin+1) ;
        if (out_tmp != NULL) {
            cpl_table_delete(out) ;
            out = out_tmp ;
        }
    }

    /* Use the model_purge flag */
    if (model_purge) {
        if (cpl_table_has_column(out, CRIRES_COL_WAVELENGTH_MODEL)) {
            cpl_table_and_selected_double(out, CRIRES_COL_WAVELENGTH_MODEL, 
                    CPL_LESS_THAN, 10.0) ;
            cpl_table_erase_selected(out) ;
        }
    }
    
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
   @brief   Extract the frames with the given tag from a frameset
   @param   in      A non-empty frameset
   @param   tag     The tag of the requested frames   
   @return  The newly created frameset or NULL on error

   The returned frameset must be de allocated with cpl_frameset_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_frameset * crires_extract_frameset(
        const cpl_frameset  *   in,
        const char          *   tag)
{
    cpl_frameset    *   out ;
    const cpl_frame *   cur_frame ;
    cpl_frame       *   loc_frame ;
    int                 nbframes, nbext ;
    int                 i ;

    /* Test entries */
    if (in == NULL) return NULL ;
    if (tag == NULL) return NULL ;

    /* Initialise */
    nbframes = cpl_frameset_get_size(in) ;

    /* Count the frames with the tag */
    if ((nbext = cpl_frameset_count_tags(in, tag)) == 0) return NULL ;

    /* Create the output frameset */
    out = cpl_frameset_new() ;

    /* Loop on the requested frames and store them in out */
    nbext = 0 ;
    for (i=0 ; i<nbframes ; i++) {
        cur_frame = cpl_frameset_get_position_const(in, i) ;
        if (!strcmp(cpl_frame_get_tag(cur_frame), tag)) {
            loc_frame = cpl_frame_duplicate(cur_frame) ;
            cpl_frameset_insert(out, loc_frame) ;
            nbext ++ ;
        }
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
   @brief   Extract the filename ﬀor the first frame of the given tag
   @param   in      A non-empty frameset
   @param   tag     The tag of the requested file 
   @return  Pointer to the file
 */
/*----------------------------------------------------------------------------*/
const char * crires_extract_filename(
        const cpl_frameset  *   in, 
        const char          *   tag) 
{
    const cpl_frame *   cur_frame ;

    /* Get the frame  */
    if ((cur_frame = cpl_frameset_find_const(in, tag)) == NULL) return NULL ;
    return cpl_frame_get_filename(cur_frame) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save a CRIRES_NB_DETECTORS extensions CRIRES image 
  @param    allframes  The list of input frames for the recipe
  @param    parlist    The list of input parameters
  @param    usedframes The list of raw/calibration frames used for this product
  @param    images     The image list to be saved
  @param    recipe     The recipe name
  @param    procat     The product category tag
  @param    protype    The product type tag
  @param    period
  @param    applist    Optional propertylist to append to primary header or NULL
  @param    applists   Optional propertylists to append to ext headers or NULL
  @param    pipe_id    PACKAGE "/" PACKAGE_VERSION
  @param    filename   Filename
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_image_save(
        cpl_frameset                *   allframes,
        const cpl_parameterlist     *   parlist,
        const cpl_frameset          *   usedframes,
        const cpl_imagelist         *   images,
        const char                  *   recipe,
        const char                  *   procat,
        const char                  *   protype,
        crires_illum_period             period,
        const cpl_propertylist      *   applist,
        const cpl_propertylist      **  applists,
        const char                  *   pipe_id, 
        const char                  *   filename)
{
    cpl_image           *   out ;
    cpl_propertylist    *   pro_list ;
    cpl_type                pixeltype ; 
    cpl_propertylist    *   qc_ext_list ;
    char                    sval[16] ;
    int                     ly ;
    cpl_type                file_type ;
    int                     i ;
    
    /* Test entries */
    if (allframes == NULL) return -1 ;
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;

    /* Initialise */
    ly = -1 ;

    /* Add the PRO keys */
    if (applist != NULL) pro_list = cpl_propertylist_duplicate(applist) ;
    else pro_list = cpl_propertylist_new() ;
    if (procat != NULL) 
        cpl_propertylist_append_string(pro_list, CPL_DFS_PRO_CATG, procat) ;
    if (protype != NULL) 
        cpl_propertylist_append_string(pro_list, CPL_DFS_PRO_TYPE, protype) ;

    /* File with extensions */
    if (cpl_dfs_save_image(allframes, NULL, parlist, usedframes, NULL, NULL,
                CPL_BPP_IEEE_FLOAT, recipe, pro_list, NULL, pipe_id, 
                filename) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the empty primary HDU") ;
        cpl_propertylist_delete(pro_list) ;
        return -1 ;
    }
    /* Delete PRO LIST */
    cpl_propertylist_delete(pro_list) ;

    /* Get the file type */
    file_type = cpl_image_get_type(cpl_imagelist_get_const(images, 0)) ;
    if (file_type == CPL_TYPE_INT)  pixeltype = CPL_BPP_32_SIGNED ;
    else                            pixeltype = CPL_BPP_IEEE_FLOAT ;
    
    /* Save the extensions */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if ((applists != NULL) && (applists[i] != NULL)) 
            qc_ext_list = cpl_propertylist_duplicate(applists[i]) ;
        else 
            qc_ext_list = cpl_propertylist_new() ;
        sprintf(sval, "CHIP%d.INT1", i+1) ;
        cpl_propertylist_prepend_string(qc_ext_list, "EXTNAME", sval) ;

        if (period == CRIRES_ILLUM_FULL_DETECTOR) {
            /* Windowing mode : the full detector is used */
            cpl_image_save(cpl_imagelist_get_const(images, i), 
                    filename, pixeltype, qc_ext_list, CPL_IO_EXTEND) ;
        } else {
            /* Non-windowing mode : use the period to know which part of the */
                /* detector is loaded / saved */
            /* Position of the image */
            if (i+1 == 1) ly = crires_get_detector_ly1(period) ;
            if (i+1 == 2) ly = crires_get_detector_ly2(period) ;
            if (i+1 == 3) ly = crires_get_detector_ly3(period) ;
            if (i+1 == 4) ly = crires_get_detector_ly4(period) ;

            /* Create the output image */
            out = cpl_image_new(1024, 512, file_type) ;
            
            /* Special case for FLAT image : non-illuminated regions set to 1 */
            if (!strcmp(procat, CRIRES_CALPRO_FLAT))
                cpl_image_add_scalar(out, 1.0) ;

            /* Copy the image to save in the output image */
            if (cpl_image_copy(out, cpl_imagelist_get_const(images, i),
                        1, ly) != CPL_ERROR_NONE) {
                cpl_msg_error(__func__, "Cannot recreate output image") ;
                cpl_image_delete(out) ;
                cpl_propertylist_delete(qc_ext_list) ;
                return -1 ;
            }

            /* Save */
            cpl_image_save(out,filename,pixeltype, qc_ext_list, CPL_IO_EXTEND) ;
            cpl_image_delete(out) ;
        }
        cpl_propertylist_delete(qc_ext_list) ;
    }
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save a CRIRES_NB_DETECTORS extensions CRIRES table
  @param    allframes  The list of input frames for the recipe
  @param    parlist    The list of input parameters
  @param    usedframes The list of raw/calibration frames used for this product
  @param    tables     The table list to be saved
  @param    recipe     The recipe name
  @param    procat     The product category tag
  @param    protype    The product type tag
  @param    applist    Optional propertylist to append to primary header or NULL
  @param    applists   Optional propertylists to append to ext headers or NULL
  @param    pipe_id    PACKAGE "/" PACKAGE_VERSION
  @param    filename   Filename
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_table_save(
        cpl_frameset                *   allframes,
        const cpl_parameterlist     *   parlist,
        const cpl_frameset          *   usedframes,
        const cpl_table             **  tables,
        const char                  *   recipe,
        const char                  *   procat,
        const char                  *   protype,
        const cpl_propertylist      *   applist,
        const cpl_propertylist      **  applists,
        const char                  *   pipe_id, 
        const char                  *   filename)
{
    cpl_propertylist    *   pro_list ;
    cpl_propertylist    *   qc_ext_list ;
    char                    sval[16] ;
    int                     i ;

    /* Test entries */
    if (allframes == NULL) return -1 ;

    /* Add the PRO keys */
    if (applist != NULL) pro_list = cpl_propertylist_duplicate(applist) ;
    else pro_list = cpl_propertylist_new() ;
    if (procat != NULL) 
        cpl_propertylist_append_string(pro_list, CPL_DFS_PRO_CATG, procat) ;
    if (protype != NULL) 
        cpl_propertylist_append_string(pro_list, CPL_DFS_PRO_TYPE, protype) ;

    /* File with extensions */
    if ((applists != NULL) && (applists[0] != NULL)) 
        qc_ext_list = cpl_propertylist_duplicate(applists[0]) ;
    else 
        qc_ext_list = cpl_propertylist_new() ;
    cpl_propertylist_prepend_string(qc_ext_list, "EXTNAME", "CHIP1.INT1") ;
    if (cpl_dfs_save_table(allframes, NULL, parlist, usedframes, NULL,
                tables[0], qc_ext_list, recipe, pro_list, NULL, 
                pipe_id, filename) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the first extension table") ;
        cpl_propertylist_delete(qc_ext_list) ;
        cpl_propertylist_delete(pro_list) ;
        return -1 ;
    }
    cpl_propertylist_delete(qc_ext_list) ;
    cpl_propertylist_delete(pro_list) ;

    /* Save the extensions */
    for (i=1 ; i<CRIRES_NB_DETECTORS ; i++) {
        if ((applists != NULL) && (applists[i] != NULL))
            qc_ext_list = cpl_propertylist_duplicate(applists[i]) ;
        else 
            qc_ext_list = cpl_propertylist_new() ;
        sprintf(sval, "CHIP%d.INT1", i+1) ;
        cpl_propertylist_prepend_string(qc_ext_list, "EXTNAME", sval) ;
        cpl_table_save(tables[i], NULL, qc_ext_list, filename, CPL_IO_EXTEND) ;
        cpl_propertylist_delete(qc_ext_list) ;
    }

    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Convert an image list of CRIRES_NB_DETECTORS images to a CRIRES 
            image
  @param    ilist   the input image list
  @return   the CRIRES_NB_DETECTORS images image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_imlist_to_image(const cpl_imagelist * ilist) 
{
    int                 nx, ny, lx, ly, gap ;
    int                 nb_det ;
    const cpl_image *   cur_im ;
    const float     *   pinf ;
    const int       *   pini ;
    cpl_image       *   out ;
    float           *   poutf ;
    int             *   pouti ;
    int                 pos1, pos2 ;
    cpl_type            pixeltype ; 
    int                 i, k, j ;
    
    /* Initialise */
    gap = 0 ;
    nb_det = CRIRES_NB_DETECTORS ;
    
    /* Test entries */
    if (ilist == NULL) return NULL ;
    if (cpl_imagelist_get_size(ilist) != nb_det) return NULL ;

    /* Output image size */
    cur_im = cpl_imagelist_get_const(ilist, 0) ;
    lx = cpl_image_get_size_x(cur_im) ;
    ly = cpl_image_get_size_y(cur_im) ;
    for (i=0 ; i<nb_det ; i++) {
        cur_im = cpl_imagelist_get_const(ilist, i) ;
        if (lx != cpl_image_get_size_x(cur_im)) return NULL ;
        if (ly != cpl_image_get_size_y(cur_im)) return NULL ;
    }
    nx = nb_det * lx + (nb_det-1) * gap ;
    ny = ly ;

    /* Get the pixel type */
    pixeltype = cpl_image_get_type(cpl_imagelist_get_const(ilist, 0)) ;
    
    /* Create the output image */
    out = cpl_image_new(nx, ny, pixeltype) ;

    /* Support FLOAT AND INT */
    if (pixeltype == CPL_TYPE_FLOAT) {
        poutf = cpl_image_get_data_float(out) ;
        /* Write the image */
        for (i=0 ; i<nb_det ; i++) {
            cur_im = cpl_imagelist_get_const(ilist, i) ;
            pinf = cpl_image_get_data_float_const(cur_im) ;
            for (j=0 ; j<lx ; j++) {
                for (k=0 ; k<ly ; k++) {
                    pos1 = j + k * lx ;
                    pos2 = (j+i*(lx+gap)) + k * nx ;
                    poutf[pos2] = pinf[pos1] ;
                }
            }
        }
    } else if (pixeltype == CPL_TYPE_INT) {
        pouti = cpl_image_get_data_int(out) ;
        /* Write the image */
        for (i=0 ; i<nb_det ; i++) {
            cur_im = cpl_imagelist_get_const(ilist, i) ;
            pini = cpl_image_get_data_int_const(cur_im) ;
            for (j=0 ; j<lx ; j++) {
                for (k=0 ; k<ly ; k++) {
                    pos1 = j + k * lx ;
                    pos2 = (j+i*(lx+gap)) + k * nx ;
                    pouti[pos2] = pini[pos1] ;
                }
            }
        }
    } else {
        cpl_image_delete(out) ;
        out = NULL ; 
    }
    return out ;
}

/**@}*/

static int crires_table_check_column(
        const cpl_table     *   tab,
        const char          *   col) 
{
    int         ret ;
    if (!cpl_table_has_column(tab, col)) {
        cpl_msg_error(__func__, "Column %s is missing", col) ;
        ret = 1 ;
    } else ret = 0 ;
    return ret ;
}

