/* $Id: crires_model_computation.c,v 1.22 2012-01-12 17:58:47 bristowp Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: bristowp $
 * $Date: 2012-01-12 17:58:47 $
 * $Revision: 1.22 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                               Includes
 -----------------------------------------------------------------------------*/
#if !defined(_XOPEN_SOURCE)
#define _XOPEN_SOURCE  /* drand48 */
#endif

#include <cpl.h>

#include "crires_dfs.h"
#include "crires_model_computation.h"
#include "crires_model_kernel.h"
#include "crires_model_io.h"
#include "crires_model_refining.h"

/*-----------------------------------------------------------------------------
                               Define
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_model_computation       Model computation
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The actual model computation takes place here
  @param    cri_config_file
  @param    wave_eg_file
  @param    order_calc_flag
  @param    order_out_flag
  @param    trace_out_flag
  @param    is_out_flag
  @param    back_map_flag
  @param    continuum
  @param    morder
  @param    flux_scale1000
  @param    spotsize1000
  @return   NULL in error case
 */
/*----------------------------------------------------------------------------*/
int crires_model_computation(
        const char*   cri_config_file,
        const char*   wave_eg_file,
        int             order_calc_flag,
        int             order_out_flag,
        int             trace_out_flag,
        int             is_out_flag,
        int             back_map_flag,
        int             continuum,
        int             morder,
        double          flux_scale,
        double          spotsize)
{
  /*store copies of the initial entrance slit position*/
  DOUBLE es_x_init,es_y_init; 
  int morder_cnt, morder_it; /* grating spectral order */
  DOUBLE lambda, lam_min, lam_max, lamstep; /*wavelength in mm */
  int prism_enc_flag=0;
  int bgflag=1;
  int RectFlag=0;
  int newslit=1;
  int custom_list_flag=0;
  int line_atlas_flag=1;
  int is_fibre_flag=0;
  DOUBLE lambda_nan; /* wavelength in angstrom */
  DOUBLE inten; /* intensity given in line list */
  /*Monte Carlo slit illumination variables*/
  DOUBLE prob, prob2, photons, gauss1, gauss2, flag;
  DOUBLE bg=1.0; /* background level */

  DOUBLE lambda_nm;
  DOUBLE lambda0, inout, lambda_min, lambda_max;
  DOUBLE xdet_min, ydet_min, xis_min, yis_min, xdet_max, ydet_max;
  DOUBLE xis_max, yis_max, pdis, order_width_is, order_width_det,lambda_slitmin;
  DOUBLE lambda_slitmax,slitw_det, xdet0, ydet0, xis0, yis0, xdet_min_is;
  DOUBLE ydet_min_is, xdet_max_is, ydet_max_is, disp_guess, chip_cen_lam_old;
  DOUBLE chip_up_lam_old, chip_down_lam_old, det_lam_sens, xdet_old,conv_test;
  DOUBLE mup1_old, psens;
  DOUBLE chip_cen_lam[4]={0.0},chip_up_lam[4]={0.0},chip_down_lam[4]={0.0};
  /*lines has wavelength, order (as float), intensity, chip no. 
    and x and y pix position*/
  DOUBLE lines[6][50000]={{0.0},{0.0}};
  int feat_den4[200]={0.0},feat_den7[200]={0.0};
  int feat_den6[200]={0.0},feat_den15[200]={0.0};
  int line_no, total_lines,ii,jj,kk,ll; /*counters for reading in line list */
  int chipx_current=0.0;/*pixel array counters*/
  int half_prof_wid; /*pixel array counters*/
  DOUBLE pixxlast=0.0; /*pixel array counters*/
  DOUBLE pixylast=0.0; /*pixel array counters*/
  DOUBLE pixwavlast=0.0; /*pixel array counters*/
  DOUBLE pos1; /*pixel array counters*/
  /*chip numbers (0-3 are real, 4-8 are off the chips) and x,y positions*/

  int totwave,totslit,enc;

  /*rectification variables*/
  //  DOUBLE xcen_frac,ycen_frac;
  DOUBLE wav_density,xpospix_last;
  DOUBLE x_spline[4], flux_spline[4];
  DOUBLE** znse_ref;
  int xcen_index,ycen_index;
  cpl_table*   lines_tab ;
  int           lines_number ;
  long naxes[2]={0};
  /*memory allocation for 1D representation of CCD array for fitsio*/
  double* oneD_AB; /* 1D array to take contents of ccd */
  //oneD_AB=(double*)cpl_malloc((size_t)(ASIZE*BSIZE)*sizeof(double));
  double** ccd=NULL; /*ccd array */
  double** ccd0=NULL; /*ccd array */
  double** ccd1=NULL; /*ccd array */
  double** ccd2=NULL; /*ccd array */
  double** ccd3=NULL; /*ccd array */
  double** ccd_wv=NULL; /*ccd array */
  double** detec=NULL; /*ccd array */
  double** rect=NULL; /*ccd array */
  cpl_image* ima4chip ;
  cpl_image* ima2 ;
  cpl_image* ima4 ;
  cpl_image* ima5 ;
  cpl_image* ima6 ;
  cpl_image* ima7 ;

  double wvlen_test;
  int wvlen_test_flag=0;
  
  /*fitsfile* fptr; file pointer */
  if ((ccd=crires_model_alloc2Darray(ASIZE, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((ccd0=crires_model_alloc2Darray(chipxpix, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((ccd1=crires_model_alloc2Darray(chipxpix, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((ccd2=crires_model_alloc2Darray(chipxpix, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((ccd3=crires_model_alloc2Darray(chipxpix, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((ccd_wv=crires_model_alloc2Darray(ASIZE, BSIZE))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }
  if ((rect=crires_model_alloc2Darray(5000, 512))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  }

  /* Initialise */
  morder_it = 0 ;
  if (wvlen_test_flag==1) {
    wvlen_test=crires_model_wvlen(detec,45,209221,37512,67.7,1,500.5,250.5);
    printf("out %lf \n",wvlen_test);
  }
  else if (custom_list_flag==1) {
    /*Below is commented out to get rid of compilation warning,
      uncomment to use*/
/*     double**  cust_lines; */
/*     int line_no=0; */
/*     FILE* setupfile; */
/*     char* setupid[10]; */
/*     setupfile=fopen("setupid.txt","r"); */
/*     while (fscanf(setupfile,"%s\n",&setupid[0]) !=EOF) { */
/*       printf("%s \n",setupid); */
/*       /\*Below is commented out to get rid of compilation waring, */
/* 	uncomment to use*\/ */
/*       if ((cust_lines=crires_model_refining_custom_lines("ThAr_metrology_scale.fits",5,"dummy_exp.fits", "metrology.fits",setupid,0))==NULL) { */
/* 	cpl_msg_error(__func__, "can't call cust lines"); */
/* 	return -1; */
/*       } */
/*       line_no=0; */
/*       while (cust_lines[0][line_no]!=0.0) { */
/* 	printf("%lf %lf %lf %lf %lf %lf \n",cust_lines[0][line_no],cust_lines[1][line_no],cust_lines[2][line_no],cust_lines[3][line_no]+1,cust_lines[4][line_no],cust_lines[5][line_no]); */
/* 	line_no++; */
/*       } */
/*       if ((crires_model_free2Darray(cust_lines,6))!=0) { */
/* 	cpl_msg_error(__func__, "Cannot free 2D array"); */
/* 	return -1; */
/*       } */
      
/*     } */
/*     fclose(setupfile); */
/*     if ((crires_model_free2Darray(cust_lines,6))!=0) { */
/*       cpl_msg_error(__func__, "Cannot free 2D array"); */
/*       return -1; */
/*     } */
/*     return 0; */
    printf("uncomment code to use this\n");
  }
  else {

  /*crires is the main structure with all the instrument configuration parameters
  Structure is defined in crires_model.h.h, *most* of the contents are filled from
   cri_config_file by the readfile routine */
  struct crires crires_model;
  struct crires* p_crires; /*pointer to crires_model */

  /* structures and arrays for annealing - in
     non-annealing version they are filled by the
     readfile routine but not used beyond that */
  ann_all_par all_par[300],* p_all_par;
  double abest[300];
  double amin[300];
  double amax[300];
  int aname[300];
  int adim;

  double *trace_xdisp=NULL; /*trace array */
  double *trace_flux=NULL; /*trace array */
  double *trace_lam=NULL; /*trace array */

  FILE* f_wav_enc;
  FILE* order_out;
  FILE* trace_out;
  FILE* cen_out;
  FILE* for_jef;
 /*pointers to input/output files */

  p_crires=&crires_model;
  p_all_par=&all_par[0];


  /*znse_ref is the array holding Sellmeier coefs for ZnSe refractive
    index at each temp. Currently hard coded to take 6 temperatures in [2][*] to
    [7][*], can be extended. First two rows ([0][*] to [1][*]) take the entries
    for the temperature bracketing the operation.
  Column [*][0] takes the temperature, the other columns, [*][1] to [*][6] take
  the Sellmeier coeffs */

  if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return -1;
  };

  /*initalize the model.the cfg file is
    read and the variables assigned to the crires structure
    initialisation. cri_init then Sets up those
    matrices that are not dependent upon wavelength */
  if ((adim=crires_model_readfits(abest,
				  amin,
				  amax,
				  aname,
				  cri_config_file,
				  p_crires,
				  p_all_par))==-1) {
    cpl_msg_error(__func__, "Failed to read config FITS file" );
    return -1;
  }


  /*open and read refractive index data file */
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    return -1 ;
  }
  inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
  printf("%lf %lf\n",inout*180./CPL_MATH_PI,
         sin(p_crires->mug-inout/2.0)+sin(p_crires->mug+inout/2.0));

  /* For a list of wavelengths and encoder values, just
     use the model kernel iteratively calculate the
     angle which places the wavelength at the centre of
     the IS, this is then the prism angle that
     corresponds to the encoder value */
  if (prism_enc_flag==1) {
    f_wav_enc=fopen("wav_enc_quick.dat","r");
    while (fscanf(f_wav_enc,"%lf %d\n",&lambda,&enc) !=EOF) {
      p_crires->mup1=37.7*DEG2RAD; // simply a guess value to start
      psens=-0.074762*DEG2RAD; //from trial and error...
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	return -1;
      }
      crires_model_p_rot(p_crires);
      /*Note, the order value passed below is arbitrary,
        we are only interested in results up to the IS*/
      crires_model_eval(lambda/1000000.0,40,znse_ref, p_crires);
      //printf("%lf, %lf %lf, %lf \n",lambda,enc,p_crires->yis,p_crires->mup1);
      yis0=p_crires->yis;
      mup1_old=p_crires->mup1;
      p_crires->mup1=p_crires->mup1-(yis0-p_crires->slity)*psens;
      for (ii=1;ii<6;ii++) {
        crires_model_p_rot(p_crires);
        crires_model_eval(lambda/1000000.0,40,znse_ref, p_crires);
        // printf("%lf, %lf %lf, %lf \n",lambda,enc,p_crires->yis,p_crires->mup1);
        psens=(p_crires->mup1-mup1_old)/(p_crires->yis-yis0);
        //printf("%lf %lf %lf %lf\n", yis,p_crires->slity,p_crires->mup1/DEG2RAD, psens);
        mup1_old=p_crires->mup1;
        yis0=p_crires->yis;
        p_crires->mup1=p_crires->mup1-(p_crires->yis-p_crires->slity)*psens;
      }
      printf("%d %lf %lf\n",enc, p_crires->mup1/DEG2RAD, lambda);
      /* All the stuff below is to check wavelength and
         slit scales at the intermediate slit and
         detector. Be careful, the lambda used should
         correspond to that in the lambda vs enc value
         file supplied
      p_crires->es_x=0.0;
      crires_model_eval((lambda+0.5)/1000000.0,40,znse_ref, p_crires);
      printf("%lf %lf\n", p_crires->yis, lambda+0.50);
      crires_model_eval((lambda-0.5)/1000000.0,40,znse_ref, p_crires);
      printf("%lf %lf\n", p_crires->yis, lambda-0.5);
      crires_model_eval((lambda)/1000000.0,45,znse_ref, p_crires);
      printf("45 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      crires_model_eval((lambda)/1000000.0,15,znse_ref, p_crires);
      printf("15 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      p_crires->es_x=-4.0;
      crires_model_eval((lambda)/1000000.0,45,znse_ref, p_crires);
      printf("45 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      p_crires->es_x=+4.0;
      crires_model_eval((lambda)/1000000.0,45,znse_ref, p_crires);
      printf("45 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      p_crires->es_x=-4.0;
      crires_model_eval((lambda)/1000000.0,15,znse_ref, p_crires);
      printf("15 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      p_crires->es_x=+4.0;
      crires_model_eval((lambda)/1000000.0,15,znse_ref, p_crires);
      printf("15 %lf %lf %lf %lf %lf\n", p_crires->xis, p_crires->yis, lambda, p_crires->xdet,p_crires->ydet);
      p_crires->es_x=0.0;*/
    }
  }
  /* Loop over orders to get a summary of upper and
     lower wavelength and spatial limits at the slit and
     on the detector */
  else if (order_calc_flag==1) {
      //loop over orders
    order_out=fopen("order_det_lim.dat","w");
    for (morder_cnt=11; morder_cnt<76; morder_cnt++) {
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	return -1;
      }
      lambda0=(sin(p_crires->mug-inout/2.0)+sin(p_crires->mug+inout/2.0))/
        ((morder_cnt*1.0)*p_crires->sg);
      lambda_min=(double)(morder_cnt)*lambda0/((double)(morder_cnt)+0.5);
      lambda_max=(double)(morder_cnt)*lambda0/((double)(morder_cnt)-0.5);
      crires_model_p_rot(p_crires);
      printf("%lf \n",p_crires->es_y);
      crires_model_eval(lambda0,morder_cnt,znse_ref, p_crires);
      xdet0=p_crires->xdet;
      ydet0=p_crires->ydet;
      xis0=p_crires->xis;
      yis0=p_crires->yis;

      crires_model_eval(lambda_min,morder_cnt,znse_ref, p_crires);
      xdet_min=p_crires->xdet;
      ydet_min=p_crires->ydet;
      xis_min=p_crires->xis;
      yis_min=p_crires->yis;

      crires_model_eval(lambda_max,morder_cnt,znse_ref, p_crires);
      xdet_max=p_crires->xdet;
      ydet_max=p_crires->ydet;
      xis_max=p_crires->xis;
      yis_max=p_crires->yis;
      //compute prism dispersion
      pdis=-(lambda_max-lambda_min)/(yis_max-yis_min);

      order_width_is=yis_min-yis_max;
      order_width_det=xdet_max-xdet_min;
      lambda_slitmin=lambda0-0.5*p_crires->slitw*pdis;
      lambda_slitmax=lambda0+0.5*p_crires->slitw*pdis;

      crires_model_eval(lambda_slitmax,morder_cnt,znse_ref, p_crires);
      xdet_min_is=p_crires->xdet;
      ydet_min_is=p_crires->ydet;

      crires_model_eval(lambda_slitmin,morder_cnt,znse_ref, p_crires);
      xdet_max_is=p_crires->xdet;
      ydet_max_is=p_crires->ydet;
      slitw_det=-(xdet_max_is-xdet_min_is);
      disp_guess=(lambda_slitmax-lambda_slitmin)/slitw_det;
      for (ii=0;ii<4;ii++) {
        chip_cen_lam[ii]=lambda0+
          (xdet0-(p_crires->chipx[ii]+p_crires->det_array_x))*disp_guess;
        crires_model_eval(lambda0,morder_cnt,znse_ref, p_crires);
        chip_cen_lam_old=lambda0;
        xdet_old=p_crires->xdet;
        conv_test=1.0;
        while (conv_test>0.00000001) {
          crires_model_eval(chip_cen_lam[ii],morder_cnt,znse_ref, p_crires);
          det_lam_sens=(chip_cen_lam[ii]-chip_cen_lam_old)/
            (p_crires->xdet-xdet_old);
          conv_test=(p_crires->xdet-xdet_old)*(p_crires->xdet-xdet_old);
          chip_cen_lam_old=chip_cen_lam[ii];
          xdet_old=p_crires->xdet;
          chip_cen_lam[ii]=chip_cen_lam[ii]-
            (p_crires->xdet-(p_crires->chipx[ii]+p_crires->det_array_x))*
            det_lam_sens;
        }
        chip_up_lam[ii]=
          chip_cen_lam[ii]-(chipxpix/2.0)*p_crires->pix*disp_guess;
        crires_model_eval(lambda0,morder_cnt,znse_ref, p_crires);
        chip_up_lam_old=lambda0;
        xdet_old=p_crires->xdet;
        conv_test=1.0;
        while (conv_test>0.00000001) {
          crires_model_eval(chip_up_lam[ii],morder_cnt,znse_ref, p_crires);
          det_lam_sens=
            (chip_up_lam[ii]-chip_up_lam_old)/(p_crires->xdet-xdet_old);
          conv_test=(p_crires->xdet-xdet_old)*(p_crires->xdet-xdet_old);
          chip_up_lam_old=chip_up_lam[ii];
          xdet_old=p_crires->xdet;
          chip_up_lam[ii]=
            chip_up_lam[ii]-
            (p_crires->xdet-(p_crires->chipx[ii]+p_crires->det_array_x)-
             (chipxpix/2.0)*p_crires->pix)*det_lam_sens;
        }
        chip_down_lam[ii]=
          chip_cen_lam[ii]+(chipxpix/2.0)*p_crires->pix*disp_guess;
        crires_model_eval(lambda0,morder_cnt,znse_ref, p_crires);
        chip_down_lam_old=lambda0;
        xdet_old=p_crires->xdet;
        conv_test=1.0;
        while (conv_test>0.00000001) {
          crires_model_eval(chip_down_lam[ii],morder_cnt,znse_ref, p_crires);
          det_lam_sens=
            (chip_down_lam[ii]-chip_down_lam_old)/(p_crires->xdet-xdet_old);
          conv_test=(p_crires->xdet-xdet_old)*(p_crires->xdet-xdet_old);
          chip_down_lam_old=chip_down_lam[ii];
          xdet_old=p_crires->xdet;
          chip_down_lam[ii]=
            chip_down_lam[ii]-
            (p_crires->xdet-(p_crires->chipx[ii]+p_crires->det_array_x)+
             (chipxpix/2.0)*p_crires->pix)*det_lam_sens;
        }
      }
      printf("Order %d, central wavelength %lfA, min wvlen %lfA, max wvlen %lfA, prism Angle %lfdeg Ref Ind %lf\n",morder_cnt,lambda0*1.0E7,lambda_min*1.0E7,lambda_max*1.0E7,p_crires->mup1/DEG2RAD,p_crires->rind);
      printf("wvlen bottom of slit %lfA, wvlen top of slit %lfA,\n",lambda_slitmin*1.0E7,lambda_slitmax*1.0E7);
      printf("order width on int slit %lfmm, cen wavlen on int slit %lfmm, int slit min %lfmm, int slit max %lfmm, \n",order_width_is,yis0,yis_min,yis_max);
      printf("prism dispersion %lfA/mm, Prism dispersion %lfA/pix\n",pdis*1.0E7,pdis*1.0E7*0.027);
      printf("central wvlen on det %lfmm, min wvlen on det %lfmm, max wvlen on  det %lfmm, order width on det %lfmm, detector dispersion %lfmm/A\n",xdet0,xdet_min,xdet_max,order_width_det,order_width_det/(lambda_min*1.0E7-lambda_max*1.0E7));
      printf("slit width det %lfmm, wvlen bottom of slit on det %lfmm, wvlen top of slit on det %lfmm\n\n",slitw_det,xdet_min_is,xdet_max_is);
      for (ii=0;ii<4;ii++) {
        printf("central wavelength for chip %d: %lf, min wavelength for chip %d: %lf, max wavelength for chip %d: %lf \n",ii, chip_cen_lam[ii]*1000.0, ii, chip_down_lam[ii]*1000.0, ii, chip_up_lam[ii]*1000.0);
        if (order_out_flag==1) {
          fprintf(order_out,"%12.10lf %12.10lf ",
                  chip_down_lam[ii], chip_up_lam[ii]);
        }
      }
      if (order_out_flag==1) {
        fprintf(order_out,"\n");
      }
      printf("\n");
    }
    fclose(order_out);
  }
  else {
  cen_out=fopen("centres.dat","w");
  for_jef=fopen("slit_off.dat","w");
  /*keep copies of the initial entrance slit coords,
    scatter will be around this point and the evaluation 
    of the central ray on the detector will always be
    from this initial position*/
  es_x_init=p_crires->es_x;
  es_y_init=p_crires->es_y;
  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    return -1;
  }

  /* Read the input FITS lines list */
  lines_tab = cpl_table_load(wave_eg_file, 1, 0) ;
  lines_number = cpl_table_get_nrow(lines_tab) ;

  if (line_atlas_flag==1) {
    FILE* setupfile;
    char* setupid[8];
    int prism_enc, grat_enc;
  char file_name[80];
  int                     i, j, xpos, ypos ;
  int setord[211]={58,58,58,58,57,57,57,57,56,56,56,56,55,55,55,55,54,54,54,54,53,53,53,53,52,52,52,52,51,51,51,51,50,50,50,50,49,49,49,49,48,48,48,48,47,47,47,47,46,46,46,46,45,45,45,45,44,44,44,44,43,43,43,43,43,43,42,42,42,42,42,42,41,41,41,41,41,41,40,40,40,40,40,40,39,39,39,39,39,39,38,38,38,38,38,38,37,37,37,37,37,37,36,36,36,36,36,36,35,35,35,35,35,35,34,34,34,34,34,34,33,33,33,33,33,33,32,32,32,32,32,32,31,31,31,31,31,31,30,30,30,30,30,30,29,29,29,29,29,29,28,28,28,28,28,28,27,27,27,27,27,27,26,26,26,26,26,26,25,25,25,25,25,25,24,24,24,24,24,24,23,23,23,23,23,23,22,22,22,22,22,22,22,22,21,21,21,21,21,21,21,21,20,20,20,20,20,20,20,20};
  double refwv[211]={968.156,971.153,980.59,983.587,985.084,988.132,997.851,1000.899,1002.615,1005.716,1015.729,1018.83,1020.784,1023.939,1034.257,1037.413,1039.625,1042.837,1053.472,1056.685,1059.177,1062.448,1073.413,1076.685,1079.48,1082.813,1094.121,1097.454,1100.579,1103.976,1115.641,1119.038,1122.523,1125.986,1138.022,1141.485,1145.362,1148.894,1161.317,1164.849,1169.153,1172.757,1185.581,1189.185,1193.957,1197.636,1210.878,1214.557,1219.84,1223.597,1237.274,1241.031,1246.874,1250.713,1264.843,1268.681,1275.139,1279.062,1293.663,1297.586,1295.168,1299.181,1314.27,1318.283,1333.373,1337.386,1325.858,1329.965,1345.563,1349.669,1365.267,1369.373,1358.051,1362.255,1378.381,1382.585,1398.712,1402.916,1391.859,1396.166,1412.841,1417.148,1433.823,1438.13,1427.408,1431.822,1449.067,1453.482,1470.727,1475.142,1464.836,1469.364,1487.201,1491.729,1509.565,1514.094,1504.298,1508.946,1527.395,1532.043,1550.493,1555.141,1545.963,1550.737,1569.823,1574.597,1593.683,1598.457,1590.023,1594.93,1614.675,1614.675,1619.582,1639.327,1644.235,1636.689,1641.737,1662.166,1667.213,1687.642,1692.69,1686.199,1691.396,1712.534,1717.731,1738.869,1744.066,1738.822,1744.177,1766.051,1771.406,1793.28,1798.635,1794.856,1800.38,1823.02,1828.543,1851.184,1856.707,1854.644,1860.346,1883.788,1889.49,1912.931,1918.633,1918.568,1924.461,1948.746,1954.638,1978.923,1984.816,1987.067,1993.163,2018.344,2024.44,2049.621,2055.718,2060.636,2066.95,2093.097,2099.412,2125.559,2131.874,2139.842,2146.391,2173.601,2180.15,2207.36,2213.909,2225.333,2232.133,2260.545,2267.346,2295.758,2302.558,2317.848,2324.921,2354.734,2361.807,2391.621,2398.693,2418.237,2425.603,2457.114,2464.481,2495.992,2503.358,2506.806,2514.491,2548.136,2555.821,2589.466,2597.151,2630.797,2638.481,2624.436,2632.468,2668.896,2676.927,2713.355,2721.386,2757.814,2765.845,2752.813,2761.223,2801.392,2809.801,2849.971,2858.38,2898.549,2906.958};
  double slitw[211]={1.079,1.065,1.021,1.013,1.037,1.047,1.08,1.085,1.094,1.077,1.023,1.011,1.034,1.04,1.062,1.068,1.087,1.087,1.087,1.084,1.078,1.046,0.938,0.908,0.897,0.866,0.759,0.73,0.767,0.791,0.873,0.896,0.901,0.887,0.84,0.826,0.839,0.835,0.822,0.818,0.832,0.829,0.818,0.815,0.822,0.814,0.785,0.777,0.799,0.805,0.824,0.83,0.833,0.821,0.775,0.763,0.772,0.769,0.757,0.753,0.773,0.769,0.733,0.723,0.699,0.703,0.723,0.713,0.726,0.73,0.734,0.726,0.754,0.756,0.727,0.72,0.695,0.691,0.722,0.714,0.698,0.694,0.678,0.673,0.702,0.697,0.678,0.673,0.655,0.651,0.679,0.674,0.655,0.65,0.633,0.63,0.654,0.65,0.639,0.636,0.623,0.615,0.646,0.639,0.604,0.595,0.565,0.568,0.585,0.582,0.596,0.599,0.611,0.611,0.628,0.629,0.631,0.631,0.632,0.624,0.652,0.643,0.607,0.598,0.563,0.561,0.581,0.578,0.568,0.565,0.556,0.552,0.572,0.567,0.545,0.54,0.518,0.516,0.534,0.531,0.522,0.519,0.51,0.508,0.524,0.522,0.51,0.507,0.495,0.493,0.508,0.505,0.491,0.487,0.473,0.47,0.485,0.483,0.475,0.473,0.464,0.462,0.477,0.474,0.463,0.46,0.449,0.446,0.46,0.457,0.447,0.445,0.434,0.432,0.445,0.442,0.432,0.429,0.419,0.417,0.43,0.429,0.424,0.422,0.417,0.415,0.435,0.432,0.422,0.42,0.41,0.407,0.399,0.397,0.419,0.416,0.41,0.409,0.402,0.401,0.395,0.395,0.416,0.415,0.415,0.416,0.416,0.417,0.418,0.416};
  int penc[211]={51792, 51562, 50839, 50607, 50476, 50233, 49459, 49221, 49087, 48865, 48151, 47929, 47775, 47545, 46794, 46565, 46394, 46165, 45405, 45179, 45025, 44837, 44205, 44019, 43886, 43737, 43234, 43086, 42911, 42732, 42119, 41941, 41767, 41606, 41046, 40885, 40702, 40545, 39996, 39839, 39646, 39493, 38951, 38798, 38601, 38459, 37947, 37805, 37585, 37436, 36894, 36745, 36526, 36393, 35904, 35770, 35545, 35416, 34937, 34808, 34876, 34745, 34307, 34191, 33751, 33633, 33960, 33840, 33382, 33261, 32812, 32704, 33003, 32881, 32455, 32344, 31924, 31820, 32087, 31975, 31566, 31460, 31057, 30959, 31197, 31091, 30702, 30602, 30218, 30127, 30335, 30236, 29868, 29775, 29410, 29322, 29504, 29411, 29054, 28964, 28613, 28537, 28684, 28599, 28285, 28206, 27891, 27810, 27941, 27859,27526, 27443, 27110, 27030, 27143, 27059, 26722, 26639, 26303, 26234, 26314, 26241, 25949, 25877, 25585, 25518, 25574, 25506, 25225, 25156, 24876, 24814, 24846, 24784, 24530, 24468, 24215, 24156, 24166, 24108, 23866, 23807, 23565, 23508, 23498, 23443, 23216, 23161, 22934, 22880, 22851, 22800, 22591, 22540, 22331, 22280, 22232, 22184, 21986, 21938, 21739, 21691, 21625, 21581, 21396, 21352, 21167, 21122, 21040, 20999, 20826, 20785, 20613, 20571, 20475, 20437, 20276, 20238, 20077, 20039, 19929, 19892, 19736, 19700, 19544, 19508, 19478, 19443, 19297, 19263, 19117, 19084, 18938, 18907, 18952, 18920, 18777, 18745, 18602, 18571, 18428, 18397, 18434, 18402, 18251, 18219, 18068, 18036, 17885, 17857};
  int genc[211]={206479, 194978, 157827, 145707, 206688, 195209, 157621, 145501, 206917, 195430, 157369, 145254, 207124, 195660, 157154, 145041, 207359, 195899, 156926, 144815, 207571, 196100, 156640, 144532, 207741, 196279, 156352, 144247, 207941, 196520, 156164, 144062, 208193, 196758, 155907, 143808, 208401, 196980, 155669, 143572, 208622, 197209, 155432, 143338, 208838, 197431, 155187, 143095, 209055, 197664, 154971, 142882, 209275, 197882, 154721, 142635, 209473, 198093, 154493, 142410, 236214, 225155, 182371, 170665, 125227, 112733, 236591, 225542, 182383, 170683, 124790, 112290, 236985, 225945, 182371, 170677, 124350, 111852, 237338, 226308, 182366, 170679, 123932, 111435, 237680, 226660, 182360, 170679, 123534, 111037, 237999, 226989, 182353, 170679, 123158, 110665, 238295, 227295, 182351, 170684, 122812, 110318, 238569, 227577, 182333, 170673, 122484, 110000, 238797, 227819, 182346, 170695, 122218, 109736, 239026, 228057, 182352, 170709, 121980, 109499, 239213, 228250, 182332, 170698, 121755, 109284, 239338, 228388, 182330, 170705, 121588, 109124, 239445, 228505, 182319, 170704, 121451, 108998, 239513, 228585, 182317, 170712, 121359, 108916, 239563, 228646, 182312, 170719, 121295, 108864, 239597, 228692, 182305, 170725, 121243, 108826, 239635, 228745, 182303, 170737, 121190, 108787, 239713, 228838, 182298, 170748, 121093, 108705, 239871, 229014, 182294, 170760, 120903, 108531, 240181, 229346, 182289, 170775, 120540, 108183, 240743, 229934, 182289, 170796, 119896, 107553, 270116, 259641, 212449, 201344, 151110, 139234, 85209 , 72355 , 272363, 261936, 213242, 202174, 150258, 138400, 82446 , 69579 , 275856, 265495, 214483, 203461, 148939, 137096, 78135 , 65234};
    setupfile=fopen("setupid.txt","r");
    ii=0;
    p_crires->temper=72.0;
    p_crires->es_x=p_crires->slit[4];
    p_crires->es_y=0.0;
    while (fscanf(setupfile,"%s\n",&setupid[0]) !=EOF) {
      p_crires->slitw=slitw[ii];
      prism_enc=penc[ii]-36000;
      grat_enc=genc[ii]-200000;
      p_crires->mup1=p_crires->penc_co1+
	p_crires->penc_co2*(prism_enc)+
	p_crires->penc_co3*(prism_enc*prism_enc);
      p_crires->mug=p_crires->genc_co1+
	p_crires->genc_co2*(grat_enc)+
	p_crires->genc_co3*(grat_enc*grat_enc);
      printf("ref wvlen %lf grat angle: %lf prism angle: %lf %d %d\n",refwv[ii], p_crires->mug/DEG2RAD,p_crires->mup1/DEG2RAD,penc[ii],genc[ii]);
      crires_model_init(p_crires);
      jj=0;
      for (ll=0 ; ll<lines_number ; ll++) {
	lambda_nan=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
	inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);

	lambda=lambda_nan*1e-6; /*lambda in mm */

	crires_model_eval(lambda,setord[ii],znse_ref, p_crires);
	crires_model_detpix(p_crires);
	if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4 && p_crires->onslit==1) {
	  printf("%s %lf %lf %lf %d %lf %lf\n",setupid, refwv[ii], lambda_nan, inten, p_crires->chippix[0]+1,p_crires->xpospix,p_crires->ypospix);
	  jj++;
	}
      }
      printf("%s Total %d \n",setupid,jj);
      printf("\n");

      p_crires->slit_on=0;
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	return -1;
      }
      /*compute wavelength range for given configuration,
	order and prism and grating setting*/
      inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
      lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(setord[ii])+1.0));
      lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(setord[ii])-1.0));
      /* Fix position across slit to centre (width) */
      p_crires->es_y=0.0;

      /* calculate lamstep*/
      lambda=(lam_max+lam_min)/2.0;
      lamstep=(lam_max-lam_min)/(40000.0);
      crires_model_eval(lambda,setord[ii],znse_ref, p_crires);
      pos1=p_crires->xdet/p_crires->pix;
      lambda+=lamstep;
      crires_model_eval(lambda,setord[ii],znse_ref, p_crires);
      lamstep=lamstep/(2.5*((p_crires->xdet/p_crires->pix)-pos1));

      if ((detec=crires_model_alloc2Darray(ASIZE, BSIZE))==NULL){
	cpl_msg_error(__func__, "Cannot allocate 2D array");
	return -1;
      }
      for (i=0 ; i<ASIZE ; i++) {
	for (j=0 ; j<BSIZE ; j++) {
	  detec[i][j]=0.0;
	}
      }

      for (p_crires->es_x=-p_crires->es_s;
	   p_crires->es_x<=p_crires->es_s;
	   p_crires->es_x+=0.025) {
	chipx_current=0;
	pixxlast=0.0;
	pixwavlast=0.0;
	/* Iterate over wavelength range */
	for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
	  crires_model_eval(lambda,setord[ii],znse_ref, p_crires);
	  crires_model_detpix(p_crires);
	  if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
	    /*reset chipx_current when starting new chip */
	    if (chipx_current>p_crires->chippix[1]+1) {
	      chipx_current=p_crires->chippix[1];
	    }
	    if (p_crires->xpospix>(double)(chipx_current)+0.5) {
	      /* Linear interpolation for value at centre of pixel */
	      xpos=p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		p_crires->chippix[1];
	      ypos=p_crires->chippix[2];
	      if (xpos>=0 && xpos<ASIZE && ypos>=0 && ypos<BSIZE) {
		detec[xpos][ypos]=(pixwavlast+(lambda-pixwavlast)*
				   ((double)(chipx_current)+0.5-pixxlast)/
				   (p_crires->xpospix-pixxlast))*1000000.0;
	      }
	      chipx_current=p_crires->chippix[1]+1;
	    }
	    pixxlast=p_crires->xpospix;
	    pixwavlast=lambda;
	  }
	}
      }
      printf("finished map\n");

      /* Save the product */
      naxes[0]=BSIZE;
      naxes[1]=ASIZE;
      sprintf(file_name,"%5.1f.fits",refwv[ii]);
      printf("%s\n",file_name);

      ii++;

      if  ((oneD_AB=crires_model_copy2D_to_1D(detec, ASIZE,BSIZE))==NULL) {
	cpl_msg_error(__func__,"crires_model_copy2D_to_1D\n");
    }
      if ((ima2 = cpl_image_wrap_double(naxes[1], naxes[0], oneD_AB))==NULL) {
	cpl_msg_error(__func__,"cpl_image_wrap_double\n");
      }
      if ((cpl_image_save(ima2, file_name, CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT))!=CPL_ERROR_NONE) {
	cpl_msg_error(__func__,"cpl_image_save\n") ;
      }
      if ((cpl_image_unwrap(ima2))==NULL) {
	printf("cpl_image_unwrap\n") ;
      }
      if ((crires_model_free2Darray(detec,ASIZE))!=0){
	cpl_msg_error(__func__, "Cannot free 2D array");
	return -1;
      }
    }
  }




  line_no=0;
  lam_min=10.0;
  lam_max=0.0;

  for (ii=0;ii<ASIZE;ii+=1) {
    for (jj=0;jj<BSIZE;jj+=1) {
      ccd[ii][jj]=0.0;
    }
  }

  if (is_fibre_flag==1) {
    FILE* setupfile;
    char* setupid[10];
    int ccdx, ccdy;
    double p_dif, sinc_arg, blaze;
    double min_wav, max_wav;
    min_wav=740.0;//1295.0;
    max_wav=760.0;//1305.0;
    setupfile=fopen("setupid.txt","r");
    ii=0;
    while (fscanf(setupfile,"%s\n",&setupid[0]) !=EOF) {
      printf("UNCOMMENT CALL TO crires_model_preset_angles TO GET RESULTS! \n");
      //crires_model_preset_angles(setupid, p_crires);
      crires_model_init(p_crires);
      /* Loop on the lines and  call the stis model routine */
      for (ll=0 ; ll<lines_number ; ll++) {
	lambda_nan=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
	inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);

	lambda=lambda_nan*1e-6; /*lambda in mm */

	znse_ref=NULL;
	if (lambda_nan>min_wav && lambda_nan<max_wav) {
	  for (morder_cnt=5;morder_cnt<100;morder_cnt++) {
	    crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
	    crires_model_detpix(p_crires);
	    if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
	      p_dif=(sin(p_crires->grat_alpha-(-p_crires->nug))+sin(p_crires->grat_beta-(-p_crires->nug)));
	      if (p_crires->grat_alpha>p_crires->grat_beta) {
		sinc_arg=(CPL_MATH_PI/(lambda*p_crires->sg))*(cos(p_crires->grat_alpha)/cos(p_crires->grat_alpha-(-p_crires->nug)))*p_dif;
		blaze=((sin(sinc_arg))/sinc_arg)*((sin(sinc_arg))/sinc_arg);
	      }
	      else {
		sinc_arg=(CPL_MATH_PI/(lambda*p_crires->sg))*(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha-(-p_crires->nug)))*p_dif;
		blaze=(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha))*(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha))*((sin(sinc_arg))/sinc_arg)*((sin(sinc_arg))/sinc_arg);
	      }
	      ccdx=p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1];
	      ccdy=p_crires->chippix[2];
	      ccd[ccdx][ccdy]+=inten*blaze*0.4;
	      if (ccdx<4095 &&ccdx>1) {
		ccd[ccdx+1][ccdy]+=inten*blaze*0.1;
		ccd[ccdx-1][ccdy]+=inten*blaze*0.1;
		ccd[ccdx][ccdy+1]+=inten*blaze*0.1;
		ccd[ccdx][ccdy-1]+=inten*blaze*0.1;
		ccd[ccdx+1][ccdy+1]+=inten*blaze*0.05;
		ccd[ccdx-1][ccdy+1]+=inten*blaze*0.05;
		ccd[ccdx+1][ccdy-1]+=inten*blaze*0.05;
		ccd[ccdx-1][ccdy-1]+=inten*blaze*0.05;
	      }
	      //printf("%d %d %lf %lf %lf %lf %d %d %lf \n",ii, morder_cnt, lambda_nan, inten, blaze, blaze*inten, p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1],p_crires->chippix[2],p_crires->mug);
	      printf("%s %d %lf %lf %d %lf %lf \n",setupid , morder_cnt, lambda_nan, blaze*inten, p_crires->chippix[0], p_crires->xpospix,p_crires->ypospix);
	    }
	  }
	}
      }
      ii++;
    }
    naxes[0]=BSIZE;
    naxes[1]=ASIZE;
    oneD_AB=crires_model_copy2D_to_1D(ccd, ASIZE,BSIZE);
    ima4chip = cpl_image_wrap_double(naxes[1], naxes[0], oneD_AB) ;
    cpl_image_save(ima4chip, "chips_out.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
    printf("----\n");
    fclose(setupfile);
    exit(1);
  }

  if (is_out_flag==1) {
    FILE* setupfile;
    char* setupid[10];
    int slit_count;
    setupfile=fopen("setupid.txt","r");
    while (fscanf(setupfile,"%s\n",&setupid[0]) !=EOF) {
      printf("UNCOMMENT CALL TO crires_model_preset_angles TO GET RESULTS! \n");
      //crires_model_preset_angles(setupid, p_crires);
      slit_count=0;
	/* Loop on the lines and  call the stis model routine */
      for (ll=0 ; ll<lines_number ; ll++) {
	lambda_nan=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
	inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);

	lambda=lambda_nan*1e-6; /*lambda in mm */

	crires_model_p_rot(p_crires);

	crires_model_eval(lambda,0,znse_ref, p_crires);
	if
	  (p_crires->yis-p_crires->slity>-p_crires->slitw/2.0 &&
	   p_crires->yis-p_crires->slity<p_crires->slitw/2.0) {
	  printf("%lf %lf %lf \n",inten,
		 lambda_nan,p_crires->yis-p_crires->slity);
	  slit_count++;
	}
      }
      printf("%d \n",slit_count);
      printf("----\n");
    }
    fclose(setupfile);
  }

  /* Loop on the lines and  call the stis model routine */
  for (ll=0 ; ll<lines_number ; ll++) {
    lambda_nan=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
    inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);

    lambda=lambda_nan*1e-6; /*lambda in mm */

    /* A noise "spectrum" is currently added to line lists by multiple
       entries with intensity -1.0
       For each a random no. of photons is assigned - gaussian
       dist. centred on 1.0, sigma=bg and bgflag is set to 1*/
    if (inten==-1.0) {
        /* FIXME : avoid drand48() */
        /* prob=drand48(); */
        /* prob2=drand48(); */
        prob = 0.2 ;
        prob2 = 0.7 ;
        prob=drand48();
        prob2=drand48();
      gauss1=sqrt(-2.0*log(1-prob))*cos(2*CPL_MATH_PI*prob2);
      inten=((gauss1+1.0)*bg);
      if (inten<0.0) inten=bg;
      bgflag=1;
    }
    else if (inten<0.0) {
      inten=-inten;
      bgflag=2;
    }
    else {
      bgflag=0;
    }
    /*Calculate the order of the given wavelength */
    //morder=(int)(0.5+(sin(65.25*CPL_MATH_PI/180.0)+sin(61.75*CPL_MATH_PI/180.0))/((p_crires->sg)*lambda));
    /*We now need to update several matrices (that were set in the
      cri_init routine outside of the current loop) for this prism angle */
    crires_model_p_rot(p_crires);

    /*Iterate over orders immediately above and below that chosen to
      allow for light from other orders arriving on detectors*/
    for (morder_cnt=morder-morder_it;
         morder_cnt<=morder+morder_it;
         morder_cnt+=1) {
      //    for (ii=0; ii<200;ii+=4) {
      //p_crires->mup1=(36.5+(double)(ii)*0.01)*CPL_MATH_PI/180.0;
      //morder_cnt=40;
      //crires_model_p_rot(p_crires);

      /*only keep an array of lines if not background, for background this would be way too big*/
      if (bgflag==0 && continuum==0) {
        lines[0][line_no]=lambda;
        lines[1][line_no]=(double)(morder_cnt);
        lines[2][line_no]=inten;
      }
      /*set centre of entrance slit as default*/
      p_crires->es_x=es_x_init;
      p_crires->es_y=es_y_init;
      /*call model once for this entrance slit position. This will be
      the result stored in the main line list. It is also used to
      establish which lines fall on one of the detector arrays and
      should be computed for the whole slit*/
/*       lambda0=(sin(p_crires->mug-inout/2.0)+ */
/*                sin(p_crires->mug+inout/2.0))/((morder_cnt*1.0)*31.6); */
/*       crires_model_eval(lambda0,morder_cnt,znse_ref, p_crires); */
/*       yis0=p_crires->yis; */
      crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
      if (is_out_flag==2) {
/* 	printf("%d %lf %lf %lf %lf %lf\n", */
/* 	       morder_cnt,lambda*1000.0,lambda0*1000.0, */
/* 	       p_crires->yis,yis0,p_crires->yis-yis0); */
	printf("%lf %lf\n",lambda*1000.0,p_crires->yis-p_crires->slity);
      }
      if (is_out_flag==3) {
        if
        (p_crires->yis-p_crires->slity>-p_crires->slitw/2.0 &&
         p_crires->yis-p_crires->slity<p_crires->slitw/2.0) {
          printf("%lf %lf %lf --\n",inten,
                 lambda*1000.0,p_crires->yis-p_crires->slity);
          feat_den15[ii]+=1;
        }
        p_crires->slity-=7.5;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        if (p_crires->yis-p_crires->slity>-p_crires->slitw/2.0 &&
         p_crires->yis-p_crires->slity<p_crires->slitw/2.0) {
            printf("%lf %lf %lf 7.5\n",p_crires->mup1*180.0/CPL_MATH_PI,
                   lambda*1000.0,p_crires->yis-p_crires->slity);
          feat_den7[ii]+=1;
        }
        p_crires->slity+=7.5;
        p_crires->slity-=8.5;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        if
        (p_crires->yis-p_crires->slity>-p_crires->slitw/2.0 &&
         p_crires->yis-p_crires->slity<p_crires->slitw/2.0) {
            printf("%lf %lf %lf 6.5\n",p_crires->mup1*180.0/CPL_MATH_PI,
                   lambda*1000.0,p_crires->yis-p_crires->slity);
          feat_den6[ii]+=1;
        }
        p_crires->slity+=8.5;
        p_crires->slity-=5.0;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        if (p_crires->yis-p_crires->slity>-p_crires->slitw/2.0 &&
         p_crires->yis-p_crires->slity<p_crires->slitw/2.0) {
            printf("%lf %lf %lf 10.0\n",p_crires->mup1*180.0/CPL_MATH_PI,
                   lambda*1000.0,p_crires->yis-p_crires->slity);
          feat_den4[ii]+=1;
        }
        p_crires->slity+=5.0;
        p_crires->slity-=15.0;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        if (p_crires->yis-p_crires->slity>-p_crires->slitw/6.0 &&
            p_crires->yis-p_crires->slity<p_crires->slitw/6.0) {
            printf("%lf %lf %lf 0\n",p_crires->mup1*180.0/CPL_MATH_PI,
                   lambda*1000.0,p_crires->yis-p_crires->slity);
        }
        p_crires->slity+=15.0;
        /*else {
          printf("%lf %lf %lf 0\n",lambda*1000.0,p_crires->mup1*180.0/CPL_MATH_PI, p_crires->yis-p_crires->slity);
          }*/
      }
      crires_model_detpix(p_crires);
      printf("%d %d %d %lf %lf %lf \n",morder_cnt, p_crires->onslit, p_crires->chippix[0],lambda*1000000.0,p_crires->xpospix,p_crires->ypospix);

/*       else if (trace_out_flag!=1) { */
/* 	printf("p_crires->mug %lf %d %lf %lf %lf %lf %lf %d %d %d \n",p_crires->mug, */
/*              morder_cnt,lambda*1000.0,p_crires->xpospix,p_crires->ypospix,p_crires->xdet,p_crires->ydet,p_crires->chippix[0],p_crires->chippix[1],p_crires->chippix[2]); */
/*       } */
      /*only keep an array of lines if not background, for background this would be way too big*/
      if (bgflag==0 && continuum==0) {
	//        if (morder_cnt==morder) {
          lines[3][line_no]=(double)(p_crires->chippix[0]);
          lines[4][line_no]=p_crires->xpospix;
          lines[5][line_no]=p_crires->ypospix;
          if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
            //fprintf(cen_out,"%d %lf %lf %d %lf %lf %lf %6.1f \n",line_no, lines[5][line_no], lines[4][line_no], (int)(lines[3][line_no]), lines[2][line_no], p_crires->es_x, lines[0][line_no]*1.0E+06, (lines[3][line_no]*1024.0)+lines[4][line_no]);
            fprintf(cen_out,"%lf %d %lf %lf %lf \n",
                    lines[0][line_no]*1.0E+06,
                    (int)(lines[1][line_no]),
                    lines[2][line_no],
                    (lines[3][line_no]*1024.0)+lines[4][line_no],
                    lines[5][line_no]);
          }
          //fprintf(for_jef,"%lf %lf \n", lines[0][line_no]*1.0E+06, p_crires->yis-p_crires->slity);
	  //        }
      }
      /*If p_crires->chippix[0] is -1 then the IS was not passed, however we have only checked the centre of the slit, it is possible, near the wavelength cut-offs imposed by the IS, that light from the edges of the slit pass, so we check for both sides*/
      if (p_crires->chippix[0]==-1) {
        p_crires->es_y=-p_crires->es_w;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        crires_model_detpix(p_crires);
      }
      if (p_crires->chippix[0]==-1) {
        p_crires->es_y=p_crires->es_w;
        crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
        crires_model_detpix(p_crires);
      }

      /*if on one of the detector arrays, simulate photons from the
        rest of the slit. NOTE 4 is allowed, as this
        just means that the central ray falls on one of
        the gaps between detectors*/

      if ((p_crires->chippix[0]>-1 && p_crires->chippix[0]<5)) {// || p_crires->chippix[0]==7 || p_crires->chippix[0]==8) {
        /*iterate over photons in spectral line*/
        for (photons=1; photons<(int)(inten*flux_scale); photons+=1) {
          /*call random number generator twice and generate two
            gaussian distributions with mean=0, sigma=1.0 => FWHM=2.35*/
          prob=drand48();
          prob2=drand48();
          gauss1=sqrt(-2.0*log(1-prob))*cos(2*CPL_MATH_PI*prob2);
          gauss2=sqrt(-2.0*log(1-prob))*sin(2*CPL_MATH_PI*prob2);
          //        printf("%Lf %Lf %Lf %Lf \n",prob,prob2,gauss1,gauss2);
          flag=0;
          /*if background then make the x distribution on entrance slit
            appropriate for background*/
          if (bgflag!=0) {
            p_crires->es_x=p_crires->es_s*(prob-0.5);
            p_crires->es_y=p_crires->es_w*(prob2-0.5);
            //            p_crires->es_x=es_x_init+(0.4*p_crires->es_s)*gauss1;
          }
          else {
            /*otherwise make the x dist. appropriate to signal*/
            //THIS IS A HACK, OFFSET NEEDS TO BE ADDED TO CFG
            p_crires->es_x=es_x_init+spotsize*gauss1;//+0.0*p_crires->es_s;
            p_crires->es_y=es_y_init+spotsize*gauss2;
            /*Set flag if photon is not on slit (pos was drawn from a
              gaussian, it could be several sigma from the centre)*/
            if (p_crires->es_x>p_crires->es_s ||
                p_crires->es_x<-p_crires->es_s) {
              flag=1;
            }
          }
          /* y dist. (same for noise and signal)*/ //NO LONGER!
            //p_crires->es_y=es_y_init+(1.25*p_crires->es_w)*gauss2;
          /*Set flag if photon is not on slit (pos was drawn from a
            gaussian, it could be several sigma from the centre)*/
          if ((2.0*p_crires->es_y>p_crires->es_w && bgflag!=1)
	      || (2.0*p_crires->es_y<-p_crires->es_w && bgflag!=1)) {
            flag=1;
          }
          p_crires->chippix[0]=-2;
          /* Send to the model the wavelength, order, temperature,
             sellmeier refractive index coefs and a pointer to the structure
             Get back the chip no. and x,y position)*/
          crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
          crires_model_detpix(p_crires);
          if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {// && flag!=1) {
        /*ccd is an array with one location for each pixel in all four
          chips plus the immaginary pixels that would fill the gaps
          between them*/
        /*add one to value at x,y co-ord and shift according to which chip*/
            ccd[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1]][p_crires->chippix[2]]+=1.0;
            switch (p_crires->chippix[0]) {
            case 0:
              ccd0[p_crires->chippix[1]][p_crires->chippix[2]]+=1.0;
              break;
            case 1:
              ccd1[p_crires->chippix[1]][p_crires->chippix[2]]+=1.0;
              break;
            case 2:
              ccd2[p_crires->chippix[1]][p_crires->chippix[2]]+=1.0;
              break;
            case 3:
              ccd3[p_crires->chippix[1]][p_crires->chippix[2]]+=1.0;
              break;
            case 4:
              break;
            }
          }
        }
      }
      /*if not background increment the line count (only signal is recorded)*/
      if (bgflag==0 && continuum==0) line_no+=1;
    }
  }
  if (is_out_flag==2) {
    for (ii=0; ii<200;ii+=4) {
      printf("%lf %d %d %d %d %d \n",
             (36.5+(double)(ii)*0.01), feat_den15[ii],feat_den7[ii],feat_den4[ii],
             feat_den6[ii], feat_den15[ii]+feat_den7[ii]+feat_den4[ii] );
    }
  }
  fclose(cen_out);
  /* close(for_jef); */
  total_lines=line_no;

  if (trace_out_flag==1) {
    trace_out=fopen("trace.dat","w");
    if ((trace_xdisp=crires_model_alloc1Darray(ASIZE))==NULL){
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return -1;
    }
    if ((trace_flux=crires_model_alloc1Darray(ASIZE))==NULL){
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return -1;
    }
    if ((trace_lam=crires_model_alloc1Darray(ASIZE))==NULL){
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return -1;
    }

    inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);
    lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
      ((p_crires->sg)*((double)(morder)+1.0));
    lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
      ((p_crires->sg)*((double)(morder)-1.0));
    lambda=(lam_max+lam_min)/2.0;
    lamstep=(lam_max-lam_min)/(40000.0);
    p_crires->es_x=0.0;
    crires_model_eval(lambda,morder,znse_ref, p_crires);
    pos1=p_crires->xdet/p_crires->pix;
    lambda+=lamstep;
    crires_model_eval(lambda,morder,znse_ref, p_crires);
    lamstep=lamstep/(2.5*((p_crires->xdet/p_crires->pix)-pos1));
    p_crires->es_y=0.0;
    p_crires->slit_on=0;
    if (crires_model_init(p_crires)!=0) {
      cpl_msg_error(__func__, "Failed to initialise model" );
      if ((crires_model_free2Darray(znse_ref,8))!=0) {
	cpl_msg_error(__func__, "Cannot free 2D array");
	return -1;
      }
      return -1;
    }
    half_prof_wid=10.0;
    chipx_current=0;
    pixxlast=0.0;
    pixylast=0.0;
    pixwavlast=1000000.0*lam_min;
    printf("%lf %lf %lf\n",lam_min*1000000.0, lam_max*1000000.0,lamstep*1000000.0);
    for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
      lambda_nm=lambda*1000000.0;
      crires_model_eval(lambda,morder,znse_ref, p_crires);
      crires_model_detpix(p_crires);
      if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
	/*reset chipx_current when starting new chip */
	if (chipx_current>p_crires->chippix[1]+1) {
	  chipx_current=p_crires->chippix[1];
	}
	//trace is 1d, 4096 long, i.e. chips are concatenated
	//printf("%lf %d %d %d %lf %lf %lf %lf %lf %lf\n",lambda_nm,p_crires->chippix[0],p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1],p_crires->chippix[2],p_crires->xpospix,p_crires->ypospix,pixxlast,pixylast,pixwavlast,ccd[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1]][p_crires->chippix[2]]);
	trace_xdisp[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		    p_crires->chippix[1]]=pixylast+(p_crires->ypospix-pixylast)*
	  ((double)(chipx_current)+0.5-pixxlast)/(p_crires->xpospix-pixxlast);
	trace_lam[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		  p_crires->chippix[1]]=pixwavlast+(lambda_nm-pixwavlast)*
	  ((double)(chipx_current)+0.5-pixxlast)/(p_crires->xpospix-pixxlast);
	trace_flux[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		   p_crires->chippix[1]]=0.0;
	for (ii=-half_prof_wid;ii<half_prof_wid+1;ii+=1) {
	  trace_flux[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		     p_crires->chippix[1]]+=
	    ccd[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		p_crires->chippix[1]][p_crires->chippix[2]+ii];
	}
	chipx_current=p_crires->chippix[1]+1;
      }
      pixxlast=p_crires->xpospix;
      pixylast=p_crires->ypospix;
      pixwavlast=lambda_nm;
    }
    for (ii=0;ii<ASIZE;ii+=1) {
      fprintf(trace_out," %lf %lf %d %lf\n", trace_lam[ii], trace_xdisp[ii], ii, trace_flux[ii]);
    }
    cpl_free(trace_xdisp);
    cpl_free(trace_lam);
    cpl_free(trace_flux);
    fclose(trace_out);
  }

  if (back_map_flag==1) {
  /*LOOP FOR BACKGROUND MAP - iterate over every pixel in lambda and y
    entrance slit pos'n*/
    morder_it=morder;
    for (morder_it=morder-1;morder_it<=morder+1;morder_it+=1) {
      ii=0;
      totwave=0;
      totslit=0;
    //(int)(0.5+(sin(65.25*CPL_MATH_PI/180.0)+sin(61.75*CPL_MATH_PI/180.0))/((p_crires->sg)*lambda));
      lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(morder)+1.0));
      lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(morder)-1.0));
      lamstep=(lam_max-lam_min)/(1000.0);
      printf("%lf %lf %lf\n",lam_min*1000.0, lam_max*1000.0,lamstep*1000.0);
      if (trace_out_flag==2) {
	trace_out=fopen("trace.dat","w");
	for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
	  crires_model_eval(lambda,morder_it,znse_ref, p_crires);
	  crires_model_detpix(p_crires);
	  fprintf(trace_out,"%d, %lf, %lf \n",
		  p_crires->chippix[0],p_crires->xpospix,p_crires->ypospix);
	}
	fclose(trace_out);
      }
      jj=0;
      for (p_crires->es_y=-p_crires->es_w/2.0;
	   p_crires->es_y<=p_crires->es_w/2.0;
	   p_crires->es_y+=0.05) {
	printf("%lf \n", p_crires->es_y);
	for (p_crires->es_x=-p_crires->es_s;
	     p_crires->es_x<=p_crires->es_s;
	     p_crires->es_x+=0.05) {
	  ii=0;
	  newslit=1;
	  for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
	    crires_model_eval(lambda,morder_it,znse_ref, p_crires);
	    crires_model_detpix(p_crires);
	    if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
          /*wavelength value going into the pixel should
            be the value for N+0.5 (since in the model
            it is defined that if N<xpospix<N+1 then
            chippix[1]=N). So we need to check values
            either side of N+0.5 and interpolate the
            value when we assign to ccd_wv below*/
	      if (newslit==1 || chipx_current>p_crires->chippix[1]+1) {
		chipx_current=p_crires->chippix[1];
		newslit=0;
	      }
	      if ((double)(p_crires->chippix[0])*(chipxpix+chipxspace)+
		  p_crires->xpospix>(double)(chipx_current)+0.5) {
		ccd_wv[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1]][p_crires->chippix[2]]=(pixwavlast+(lambda-pixwavlast)*((double)(chipx_current)+0.5-pixxlast)/(p_crires->xpospix-pixxlast))*10000000.0;
		chipx_current=p_crires->chippix[1]+1;
            //printf("%d %d %d %lf \n",chipx_current, p_crires->chippix[1], p_crires->chippix[0],lambda);
	      }
	      pixxlast=p_crires->xpospix;
	      pixwavlast=lambda;
	      if (RectFlag==1) {
          /*The rectified image pixel then needs to be an
          interpolation of the corresponding pixels in
          ccd modified according to the local density of 
          rect pixels on the ccd grid to conserve flux*/

          /*interpolation in wavelength*/
          /*Assumptions: Features can be fitted with a
            spline; We will take 4 grid points along the
            wavelength direction to determine the
            spline; interpolation along the spatial
            direction can be done seperately (and
            linearly)*/

          /*The four points are
            int(p_crires->xpospix-0.5),-1,+1,+2
            The -0.5 arises because we are defining
            xcen_index as the second of the 4 pixels
            (which bracket a length of 3 pixels) and is
            on average 0.5 pix before the mid point*/

		xcen_index=
		  (p_crires->chippix[0]*(int)(chipxpix+chipxspace))+
		  (int)(p_crires->xpospix-0.5);
          //xcen_frac=p_crires->xpospix-(double)((int)(p_crires->xpospix-0.5));
		ycen_index=(int)(p_crires->ypospix-0.5);
          //ycen_frac=p_crires->ypospix-(double)((int)(p_crires->ypospix-0.5));

		x_spline[0]=(double)(xcen_index-1);
		x_spline[1]=(double)(xcen_index);
		x_spline[2]=(double)(xcen_index+1);
		x_spline[3]=(double)(xcen_index+2);

		flux_spline[0]=ccd[xcen_index-1][ycen_index];
		flux_spline[1]=ccd[xcen_index][ycen_index];
		flux_spline[2]=ccd[xcen_index+1][ycen_index];
		flux_spline[3]=ccd[xcen_index+2][ycen_index];

          /*Call GSL spline fit*/
          /*gsl_interp_accel *acc 
            = gsl_interp_accel_alloc ();
          gsl_spline *spline 
            = gsl_spline_alloc (gsl_interp_cspline, 4);

          gsl_spline_init (spline, x_spline, flux_spline, 4);

          evaluate spline at the wavelength/slit location

          rect[ii][jj]=gsl_spline_eval (spline, ((double)(p_crires->chippix[0])*(chipxpix+chipxspace))+p_crires->xpospix-0.5, acc);*/

          /*Now need to correct for local density of
            wavelength points on the pixel grid (only in
            the wavelength direction for the moment)*/
		wav_density=1.0/(p_crires->xpospix-xpospix_last);
		xpospix_last=p_crires->xpospix;
		rect[ii][jj]=rect[ii][jj]*wav_density;

		if (rect[ii][jj]>0.0001) {
		  printf(" %d %d %d %lf  %lf  %lf  %lf  %lf  %lf  %lf  %lf  %lf  %lf %lf \n", ii,jj,xcen_index,flux_spline[0],flux_spline[1],flux_spline[2],flux_spline[3],x_spline[0],x_spline[1],x_spline[2],x_spline[3], ((double)(p_crires->chippix[0])*(chipxpix+chipxspace))+p_crires->xpospix-0.5, rect[ii][jj], rect[ii][jj]/wav_density);
		}
		else {
		  rect[ii][jj]=0.0;
		}

          /*          gsl_spline_free (spline);
                  gsl_interp_accel_free (acc);*/

          /* Below is a linear interp for the central 2 of the 4 pts */
          //rect[ii][jj]=ccd[xcen_index][ycen_index]+(ccd[xcen_index+1][ycen_index]-ccd[xcen_index][ycen_index])*xcen_frac;
	      }
	    }
	    ii++;
	  }
	  jj++;
	}
      }
      totwave=ii;
      totslit=jj;
      printf("%d %d \n",totwave, totslit);
    }
  }

  cpl_table_delete(lines_tab) ;

  /*Mark the spaces between the chips with pixels set at 1000*/
  for (ii=0; ii<=chipypix; ii++) {
    for (jj=1;jj<=3;jj++) {
      for
      (kk=jj*(int)(chipxpix)+(jj-1)*(int)(chipxspace);
       kk<jj*(int)(chipxpix)+(jj)*(int)(chipxspace);
       kk+=1){
        ccd[kk][ii]=1000;
        //ccd_wv[kk][ii]=1000;
      }
    }
  }

  /* To rectify the simulated image (of course this
     essentially gives you back the line list and slit
     positions that you put in, its just a test 

  if (RectFlag==1) {
    for (ii=0; ii<totwave; ii++) {
      for (jj=1;jj<totslit;jj++) {
        printf("%d %d %d %d %d\n",ii,jj,waveslit[ii][jj][0],waveslit[ii][jj][1],waveslit[ii][jj][2]);
        if (waveslit[ii][jj][0]>-1 && waveslit[ii][jj][1]>-1) {
        rect[ii][jj]+=ccd_wv[waveslit[ii][jj][0]*750+waveslit[ii][jj][1]][waveslit[ii][jj][2]];
        }
      }
    }
  }*/

  /* set fitsio parameters for fits output */
  naxes[0]=BSIZE;
  naxes[1]=ASIZE;

  /* If you want to save an image */
  oneD_AB=crires_model_copy2D_to_1D(ccd0, chipxpix,BSIZE);
  ima4 = cpl_image_wrap_double(chipxpix, naxes[0], oneD_AB) ;
  cpl_image_save(ima4, "chip0_out.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
  cpl_image_unwrap(ima4) ;

  /* If you want to save an image */
  oneD_AB=crires_model_copy2D_to_1D(ccd1, chipxpix,BSIZE);
  ima5 = cpl_image_wrap_double(chipxpix, naxes[0], oneD_AB) ;
  cpl_image_save(ima5, "chip1_out.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
  cpl_image_unwrap(ima5) ;

  /* If you want to save an image */
  oneD_AB=crires_model_copy2D_to_1D(ccd2, chipxpix,BSIZE);
  ima6 = cpl_image_wrap_double(chipxpix, naxes[0], oneD_AB) ;
  cpl_image_save(ima6, "chip2_out.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
  cpl_image_unwrap(ima6) ;

  /* If you want to save an image */
  oneD_AB=crires_model_copy2D_to_1D(ccd3, chipxpix,BSIZE);
  ima7 = cpl_image_wrap_double(chipxpix, naxes[0], oneD_AB) ;
  cpl_image_save(ima7, "chip3_out.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
  cpl_image_unwrap(ima7) ;

  /* If you want to save an image */
  if (back_map_flag==1) {
  oneD_AB=crires_model_copy2D_to_1D(ccd_wv, ASIZE,BSIZE);
  ima2 = cpl_image_wrap_double(naxes[1], naxes[0], oneD_AB) ;
  cpl_image_save(ima2, "chips_wv.fits", CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT) ;
  cpl_image_unwrap(ima2) ;
  }

  if ((crires_model_free2Darray(ccd,ASIZE))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(ccd0,chipxpix))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(ccd1,chipxpix))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(ccd2,chipxpix))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(ccd3,chipxpix))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(ccd_wv,ASIZE))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }
  if ((crires_model_free2Darray(rect,5000))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return -1;
  }

  }
  }
  return(0);

}


/**@}*/

