/* $Id: crires_wlcalib.c,v 1.58 2012-10-09 08:18:01 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-10-09 08:18:01 $
 * $Revision: 1.58 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_wlxcorr.h"
#include "irplib_ppm.h"

#include "crires_wlcalib.h"
#include "crires_wlestimate.h"
#include "crires_combine.h"
#include "crires_extract.h"
#include "crires_dfs.h"
#include "crires_utils.h"
#include "crires_pfits.h"
#include "crires_load.h"

static int crires_wlcalib_get_2_closest(const cpl_table *, int, int *, int *) ;
static cpl_polynomial * crires_wlcalib_gen_wlpoly(const cpl_table *, int) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_wlcalib     Wavelength calibration routines
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration with a lamp
  @param    lamp        the lamp image
  @param    period
  @param    cat         the catalog
  @param    log_flag    flag to use the logarithm of the catalog
  @param    flat        the flat or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bpm or NULL
  @param    detlin      the detlin coefficients or NULL
  @param    wstart      Start wavelength at 1st pixel (or -1 if unknown)
  @param    wstop       Stop wavelength at last pixel (or -1 if unknown)
  @param    wmin        Minimum wavel. to consider (<0 if none)
  @param    wmax        Maximum wavel. to consider (<0 if none)
  @param    chip        the chip number
  @param    ypos        The Ypos list for the wave cal (start at 1)
  @param    width       The spectra width to extract for the calibration
  @param    degree      the requested degree
  @param    slitw       the slit width
  @param    fwhm        the fwhm
  @param    error       the wavelength error (in pixels)
  @param    samples     the number of samples
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    ppm         flag to use PPM
  @param    display     flag for display mode 
  @param    lines_flux  the lamp lines flux
  @param    fwhm_med    the lamp lines fwhm 
  @param    rpower      the lamp lines resolving power
  @return   the result table or NULL in error case

  The returned table contains one row per wavelength calibration.

  The following columns are contained in the table.
    - CRIRES_COL_WL_COEFFA
    - CRIRES_COL_WL_COEFFB
    - CRIRES_COL_WL_COEFFC
    - CRIRES_COL_WL_COEFFD
    - CRIRES_COL_WL_COEFFE
    - CRIRES_COL_WL_POSY   
    - CRIRES_COL_WL_XCORR 
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_wlcalib_lamp(
        const char          *   lamp,
        crires_illum_period     period,
        const char          *   cat,
        int                     log_flag,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        double                  wstart,
        double                  wstop,
        double                  wmin,
        double                  wmax,
        int                     chip,
        cpl_vector          *   ypos,
        int                     width,
        int                     degree,
        double                  slitw,
        double                  fwhm,
        double                  error,
        int                     samples,
        int                     clean_spec,
        double                  xc_limit,
        int                     ppm,
        int                     display,
        double              *   lines_flux,
        double              *   fwhm_med,
        double              *   rpower)
{
    cpl_polynomial      *   wl_est_pol ;
    cpl_table           *   wl_est_tab ;
    cpl_table           *   catal ;
    cpl_bivector        *   catal_biv ;
    double              *   catal_biv_x ;
    double              *   catal_biv_y ;
    double                  val, dit ;
    cpl_imagelist       *   lamplist ;
    cpl_imagelist       *   lamp_ima ;
    cpl_image           *   tmp_ima ;
    cpl_propertylist    *   plist ;
    cpl_vector          *   extr ;
    cpl_vector          *   ypos_corr ;
    int                     nrows, ly, ypos_nb, ny, start_y, stop_y ;
    double              *   xcorr_list ;
    cpl_polynomial      **  wl_sol_list ;
    cpl_table           *   wl_sol_tab ;
    int                     i ;
    
    /* Test entries */
    if (lamp == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (ypos == NULL) return NULL ;
    if (lines_flux == NULL || fwhm_med == NULL || rpower == NULL) return NULL ;

    /* Initialise */
    ly = -1 ;
    ypos_nb = cpl_vector_get_size(ypos) ;

    /* Compute the Wavelength Solution Estimation */
    if (wstart > 0 && wstop > 0) 
        wl_est_pol = crires_wlestimate_compute(wstart, wstop) ;
    else 
        wl_est_pol = crires_wlestimate_get(lamp, chip) ;

    /* Create the Wavelength Solution Estimation Table */
    wl_est_tab = crires_wlcalib_gen_wltab_one(wl_est_pol, -1, -1.0) ;

    /* No catalog, use the estimate */
    if (cat == NULL) {
        cpl_msg_error(__func__, "The catalog is missing") ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }
    
    /* Load the  catalog */
    catal = crires_load_table_check(cat, 1, CRIRES_PROTYPE_CATALOG, -1, -1, 0) ;
    if (catal == NULL) {
        cpl_msg_info(__func__, "The catalog cannot be loaded") ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }
    
    /* Create the catalog spectrum in a bivector */
    nrows = cpl_table_get_nrow(catal) ;
    catal_biv = cpl_bivector_new(nrows) ;
    catal_biv_x = cpl_bivector_get_x_data(catal_biv) ;
    catal_biv_y = cpl_bivector_get_y_data(catal_biv) ;
    for (i=0 ; i<nrows ; i++) {
        catal_biv_x[i] = cpl_table_get(catal, CRIRES_COL_WAVELENGTH, i, NULL) ;
        val = cpl_table_get(catal, CRIRES_COL_EMISSION, i, NULL) ;
        if (log_flag && val > 0)    catal_biv_y[i] = log10(val) ;
        else                        catal_biv_y[i] = val ;
    }
    cpl_table_delete(catal) ;

    /* Load the lamp */
    if ((lamplist = crires_load_file(lamp, period, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the lamp") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }

    /* Get the DIT value from the header */
    if ((plist=cpl_propertylist_load(lamp, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        cpl_bivector_delete(catal_biv) ;
        cpl_polynomial_delete(wl_est_pol) ;
        cpl_imagelist_delete(lamplist) ;
        return wl_est_tab ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_polynomial_delete(wl_est_pol) ;
        cpl_imagelist_delete(lamplist) ;
        return wl_est_tab ;
    }
    cpl_table_delete(wl_est_tab) ;
    
    /* Apply the calibration */
    lamp_ima = cpl_imagelist_new() ;
    tmp_ima = cpl_image_duplicate(cpl_imagelist_get(lamplist, chip-1)) ;
    ny = (int)cpl_image_get_size_y(tmp_ima) ;
    cpl_imagelist_set(lamp_ima, tmp_ima, 0) ;
    crires_calib_chip_list(lamp_ima, period, chip, flat, dark, bpm, detlin, 
            dit, -1, -1, -1) ;
    cpl_imagelist_delete(lamplist) ;

    /* Ypos reference position correction */
    /* Get the correction for the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;

    ypos_corr = cpl_vector_duplicate(ypos) ;
    if (ly > 0) cpl_vector_subtract_scalar(ypos_corr, (double)(ly-1)) ;

    /* Create holder for solutions */
    xcorr_list = cpl_malloc(ypos_nb * sizeof(double)) ;
    wl_sol_list = cpl_malloc(ypos_nb * sizeof(cpl_polynomial*)) ;

    /* Loop on the calibration Y positions */
    for (i=0 ; i<ypos_nb ; i++) {
        /* Define the extraction zone */
        start_y = cpl_vector_get(ypos_corr, i) - (int)(width/2.0) ;
        stop_y = cpl_vector_get(ypos_corr, i) + (int)(width/2.0) ;
        if (start_y < 1) start_y = 1 ;
        if (stop_y > ny) stop_y = ny ;

        /* Extract the spectrum to calibrate */
        cpl_msg_info(__func__, "Calibration Spectrum at Y=%g",
                cpl_vector_get(ypos, i)) ;
        cpl_msg_indent_more() ;
        if ((extr=crires_extract_lines(cpl_imagelist_get(lamp_ima, 0),
                        start_y, stop_y, 0, 0)) == NULL) {
            cpl_msg_warning(__func__, "Cannot extract the lamp") ;
            xcorr_list[i] = -1.0 ;
            wl_sol_list[i] = cpl_polynomial_duplicate(wl_est_pol) ;
        } else {
            /* Calibration */
            if ((wl_sol_list[i] = crires_wlcalib_engine(extr, catal_biv, 
                            wl_est_pol, slitw, fwhm, degree, wmin, wmax, 
                            error, samples, clean_spec, xc_limit, ppm, 
                            display, 0, 0, &(xcorr_list[i]), NULL)) == NULL) {
                cpl_msg_error(__func__, "Cannot compute the wavelength") ;
                xcorr_list[i] = -1.0 ;
                wl_sol_list[i] = cpl_polynomial_duplicate(wl_est_pol) ;
            } else {
                /* Compute the QC for the lamp lines */
                cpl_msg_info(__func__, "Compute the Lamp qc parameters") ;
                if (crires_extract_lines_properties(extr, wl_sol_list[i], 
                            display, lines_flux, fwhm_med, rpower) == -1) {
                    cpl_msg_warning(__func__, "Cannot compute the THAR QCs") ;
                }
            }
            cpl_vector_delete(extr) ;

            /* Plot if requested */
            if (display) {
                irplib_wlxcorr_plot_solution(wl_est_pol, wl_sol_list[i], NULL, 
                        1, 1024);
            }
        }
        cpl_msg_indent_less() ;
    }
    cpl_bivector_delete(catal_biv) ;
    cpl_imagelist_delete(lamp_ima) ;
    cpl_polynomial_delete(wl_est_pol) ;
    

    /* Corrected Y positions are shifted -> Full detector Y position */
    /* After that, ypos_corr is not any more 'corrected' */
    if (ly > 0) cpl_vector_add_scalar(ypos_corr, (double)(ly-1)) ;

    /* Create the Output table */
    wl_sol_tab = crires_wlcalib_gen_wltab_several(
            (const cpl_polynomial **)wl_sol_list,
            (const cpl_vector *)ypos_corr, 
            (const double *)xcorr_list) ;

    /* Free  */
    cpl_free(xcorr_list) ;
    for (i=0 ; i<ypos_nb ; i++) 
        cpl_polynomial_delete(wl_sol_list[i]) ;
    cpl_free(wl_sol_list) ;
    cpl_vector_delete(ypos_corr) ;

    return wl_sol_tab ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration with a sky
  @param    sky         the sky image
  @param    period
  @param    oh          the oh catalog
  @param    hitran      the hitran catalog
  @param    log_flag    flag to use the logarithm of the catalog
  @param    flat        the flat or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bpm or NULL
  @param    detlin      the detlin coefficients or NULL
  @param    wstart      Start wavelength at 1st pixel (or -1 if unknown)
  @param    wstop       Stop wavelength at last pixel (or -1 if unknown)
  @param    wmin        Minimum wavel. to consider (<0 if none)
  @param    wmax        Maximum wavel. to consider (<0 if none)
  @param    chip        the chip number
  @param    ypos        The Ypos list for the wave cal (start at 1)
  @param    width       The spectra width to extract for the calibration
  @param    degree      the requested degree
  @param    slitw       the slit width
  @param    fwhm        the fwhm
  @param    error       the wavelength error (in pixels)
  @param    samples     the number of samples
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    ppm         flag to use PPM
  @param    display     flag for display mode 
  @return   the result table or NULL in error case

  The returned table contains one row per wavelength calibration.

  The following columns are contained in the table.
    - CRIRES_COL_WL_COEFFA
    - CRIRES_COL_WL_COEFFB
    - CRIRES_COL_WL_COEFFC
    - CRIRES_COL_WL_COEFFD
    - CRIRES_COL_WL_COEFFE
    - CRIRES_COL_WL_POSY   
    - CRIRES_COL_WL_XCORR  
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_wlcalib_sky(
        const char          *   sky,
        crires_illum_period     period,
        const char          *   oh,
        const char          *   hitran,
        int                     log_flag,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        double                  wstart,
        double                  wstop,
        double                  wmin,
        double                  wmax,
        int                     chip,
        cpl_vector          *   ypos,
        int                     width,
        int                     degree,
        double                  slitw,
        double                  fwhm,
        double                  error,
        int                     samples,
        int                     clean_spec,
        double                  xc_limit,
        int                     ppm,
        int                     display)
{
    double                  wlen ;
    cpl_polynomial      *   wl_est_pol ;
    cpl_table           *   wl_est_tab ;
    cpl_table           *   catal ;
    cpl_bivector        *   catal_biv ;
    double              *   catal_biv_x ;
    double              *   catal_biv_y ;
    double                  val, dit ;
    cpl_imagelist       *   skylist ;
    cpl_imagelist       *   sky_ima ;
    cpl_image           *   tmp_ima ;
    cpl_propertylist    *   plist ;
    cpl_vector          *   extr ;
    cpl_vector          *   ypos_corr ;
    int                     nrows, ly, ypos_nb, ny, start_y, stop_y ;
    double              *   xcorr_list ;
    cpl_polynomial      **  wl_sol_list ;
    cpl_table           *   wl_sol_tab ;
    int                     i ;
    
    /* Test entries */
    if (sky == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (ypos == NULL) return NULL ;
    
    /* Initialise */
    ly = -1 ;
    ypos_nb = cpl_vector_get_size(ypos) ;
    
    /* Compute the Wavelength Solution Estimation */
    if (wstart > 0 && wstop > 0) 
        wl_est_pol = crires_wlestimate_compute(wstart, wstop) ;
    else 
        wl_est_pol = crires_wlestimate_get(sky, chip) ;

    /* Compute the estimated Central Wavelength */
    wlen = cpl_polynomial_eval_1d(wl_est_pol, 512.0, NULL) ;

    /* Create the Wavelength Solution Estimation Table */
    wl_est_tab = crires_wlcalib_gen_wltab_one(wl_est_pol, -1, -1.0) ;

    /* No catalog, use the estimate */
    if (oh == NULL && hitran == NULL) {
        cpl_msg_error(__func__, "The catalogs are missing") ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }
    
    /* Load the relevant catalog - Check the central wavelength */
    if (wlen < 1800) {
        if (oh == NULL) {
            cpl_msg_error(__func__, "Need the OH catalog below 1.8 microns") ;
            cpl_polynomial_delete(wl_est_pol) ;
            return wl_est_tab ;
        }
        catal = crires_load_table_check(oh, 1, CRIRES_PROTYPE_CATALOG, -1,-1,0);
    } else {
        if (hitran == NULL) {
            cpl_msg_error(__func__,"Need the HITRAN catalog above 1.8 microns");
            cpl_polynomial_delete(wl_est_pol) ;
            return wl_est_tab ;
        }
        catal = crires_load_table_check(hitran, 1, CRIRES_PROTYPE_CATALOG, 
                -1, -1, 0) ;
    }
    
    /* Create the catalog spectrum in a bivector */
    nrows = cpl_table_get_nrow(catal) ;
    catal_biv = cpl_bivector_new(nrows) ;
    catal_biv_x = cpl_bivector_get_x_data(catal_biv) ;
    catal_biv_y = cpl_bivector_get_y_data(catal_biv) ;
    for (i=0 ; i<nrows ; i++) {
        catal_biv_x[i] = cpl_table_get(catal, CRIRES_COL_WAVELENGTH, i, NULL) ;
        val = cpl_table_get(catal, CRIRES_COL_EMISSION, i, NULL) ;
        if (log_flag && val > 0)    catal_biv_y[i] = log10(val) ;
        else                        catal_biv_y[i] = val ;
    }
    cpl_table_delete(catal) ;

    /* Load the sky */
    if ((skylist = crires_load_file(sky, period, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the sky") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }

    /* Get the DIT value from the header */
    if ((plist=cpl_propertylist_load(sky, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(skylist) ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(skylist) ;
        cpl_polynomial_delete(wl_est_pol) ;
        return wl_est_tab ;
    }
    cpl_table_delete(wl_est_tab) ;

    /* Apply the calibration */
    sky_ima = cpl_imagelist_new() ;
    tmp_ima = cpl_image_duplicate(cpl_imagelist_get(skylist, chip-1)) ;
    ny = (int)cpl_image_get_size_y(tmp_ima) ;
    cpl_imagelist_set(sky_ima, tmp_ima, 0) ;
    crires_calib_chip_list(sky_ima, period, chip, flat, dark, bpm, detlin, dit,
            -1, -1, -1) ;
    cpl_imagelist_delete(skylist) ;

    /* Ypos reference position correction */
    /* Get the correction for the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;
    
    ypos_corr = cpl_vector_duplicate(ypos) ;
    if (ly > 0) cpl_vector_subtract_scalar(ypos_corr, (double)(ly-1)) ;

    /* Create holder for solutions */
    xcorr_list = cpl_malloc(ypos_nb * sizeof(double)) ;
    wl_sol_list = cpl_malloc(ypos_nb * sizeof(cpl_polynomial*)) ;
    
    /* Loop on the calibration Y positions */
    for (i=0 ; i<ypos_nb ; i++) {
        /* Define the extraction zone */
        start_y = cpl_vector_get(ypos_corr, i) - (int)(width/2.0) ;
        stop_y = cpl_vector_get(ypos_corr, i) + (int)(width/2.0) ;
        if (start_y < 1) start_y = 1 ;
        if (stop_y > ny) stop_y = ny ;
        
        /* Extract the spectrum to calibrate */
        cpl_msg_info(__func__, "Calibration Spectrum at Y=%g",
                cpl_vector_get(ypos, i)) ;
        cpl_msg_indent_more() ;
        if ((extr=crires_extract_lines(cpl_imagelist_get(sky_ima, 0),
                        start_y, stop_y, 0, 0)) == NULL) {
            cpl_msg_warning(__func__, "Cannot extract the sky") ;
            xcorr_list[i] = -1.0 ;
            wl_sol_list[i] = cpl_polynomial_duplicate(wl_est_pol) ;
        } else {
            /* Calibration */
            if ((wl_sol_list[i] = crires_wlcalib_engine(extr, catal_biv, 
                            wl_est_pol, slitw, fwhm, degree, wmin, wmax, 
                            error, samples, clean_spec, xc_limit, ppm, 
                            display, 0, 0, &(xcorr_list[i]), NULL)) == NULL) {
                cpl_msg_warning(__func__, "Cannot compute the wavelength") ;
                xcorr_list[i] = -1.0 ;
                wl_sol_list[i] = cpl_polynomial_duplicate(wl_est_pol) ;
            } 
            cpl_vector_delete(extr) ;
            
            /* Plot if requested */
            if (display) {
                irplib_wlxcorr_plot_solution(wl_est_pol, wl_sol_list[i], NULL, 
                        1, 1024);
            }
        }
        cpl_msg_indent_less() ;
    }
    cpl_bivector_delete(catal_biv) ;
    cpl_polynomial_delete(wl_est_pol) ;
    cpl_imagelist_delete(sky_ima) ;

    /* Corrected Y positions are shifted -> Full detector Y position */
    /* After that, ypos_corr is not any more 'corrected' */
    if (ly > 0) cpl_vector_add_scalar(ypos_corr, (double)(ly-1)) ;

    /* Create the Output table */
    wl_sol_tab = crires_wlcalib_gen_wltab_several(
            (const cpl_polynomial **)wl_sol_list,
            (const cpl_vector *)ypos_corr, 
            (const double *)xcorr_list) ;

    /* Free  */
    cpl_free(xcorr_list) ;
    for (i=0 ; i<ypos_nb ; i++) 
        cpl_polynomial_delete(wl_sol_list[i]) ;
    cpl_free(wl_sol_list) ;
    cpl_vector_delete(ypos_corr) ;

    return wl_sol_tab ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration using a spectrum
  @param    spec            the spectrum to calibrate
  @param    catalog_spec    the spectrum obtained with the catalog lines
  @param    phdisp          the physical model polynomial
  @param    slit_width      Width in pixels of the slit used
  @param    fwhm
  @param    degree          Degree of the fitted polynomial
  @param    wl_min          The Minimum wavelength to consider
  @param    wl_max          The Maximum wavelength to consider
  @param    wl_err          The wl error (in pixels)
  @param    nsamples        Number of samples for wavelength computation
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    use_ppm         Use the point pattern matching at the end
  @param    display         flag to display intermediate results
  @param    fplotted        first plotted line
  @param    lplotted        last plotted line
  @param    xcorr           the cross correlation factor (computed)
  @return   the wavelength polynomial or NULL in error case

  The returned polynomial has to be deallocated with cpl_polynomial_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlcalib_engine(
        const cpl_vector        *   spec,
        const cpl_bivector      *   catalog_spec,
        const cpl_polynomial    *   phdisp,
        double                      slit_width,
        double                      fwhm,
        int                         degree,
        double                      wl_min,
        double                      wl_max,
        double                      wl_err,
        int                         nsamples,
        int                         clean_spec,
        double                      xc_limit,
        int                         use_ppm,
        int                         display,
        int                         fplotted,
        int                         lplotted,
        double                  *   xcorr,
        cpl_table               **  wlres)
{
    double                  wl_err_wl ;
    double                  wl_start, wl_stop ;
    int                     pix_start, pix_stop, pix_min, pix_max ;
    cpl_vector          *   spec_extract ;
    cpl_vector          *   spec_clean ;
    double              *   pspec ;
    int                     filt_size ;
    double                  xc ;
    cpl_vector          *   wl_errors ;
    cpl_polynomial      *   xcdisp ;
    cpl_polynomial      *   xcdisp_tmp ;
    cpl_polynomial      *   wl_poly_ppm ;
    cpl_vector          *   filtered ;
    double                  threshold, median, stdev ;
    cpl_bivector        *   catal_loc ;
    double              *   catal_loc_wl ;
    double              *   catal_loc_intens ;
    int                     nsam_factor, wlerr_factor ;
    cpl_table           *   spc_table ;
    cpl_vector          *   xcorrs ;
    cpl_table           *   wlres_loc ;
    int                     i ;

    /* Test entries */
    if (spec == NULL || catalog_spec == NULL || phdisp == NULL) return NULL ;
    if (degree < 1 || degree > 4) return NULL ;
    if (nsamples < 10)  nsam_factor = 1 ;
    else                nsam_factor = 10 ;
    wlerr_factor = 5 ;

    /* Initialise */
    filt_size = 9 ;
    if (wlres != NULL) *wlres = NULL ;

    /* Convert the wavelength error from pixels to wavelength */
    wl_err_wl = cpl_polynomial_eval_1d(phdisp, wl_err, NULL) ;
    wl_err_wl -= cpl_polynomial_eval_1d(phdisp, 0.0, NULL) ;
    cpl_msg_info(__func__, "Wavelength error : %g pix = %g nm", wl_err,
            wl_err_wl) ;

    /* Define boundaries */
    pix_start = 1 ;
    pix_stop = cpl_vector_get_size(spec) ;
    wl_start = cpl_polynomial_eval_1d(phdisp, (double)pix_start, NULL) ;
    wl_stop = cpl_polynomial_eval_1d(phdisp, (double)pix_stop, NULL) ;
    if ((wl_min > 0) && (wl_min > wl_start)) {
        pix_min = pix_stop - 
            (int)((pix_stop-pix_start)*(wl_stop-wl_min)/(wl_stop-wl_start)) ;
        cpl_msg_warning(__func__, "Start on pixel %d", pix_min) ;
    } else pix_min = pix_start ;
    if ((wl_max > 0) && (wl_max < wl_stop)) {
        pix_max = pix_start + 
            (int)((pix_stop-pix_start)*(wl_max-wl_start)/(wl_stop-wl_start)) ;
        cpl_msg_warning(__func__, "Stop on pixel %d", pix_max) ;
    } else pix_max = pix_stop ;
    /* Check */
    if (pix_max <= pix_min) {
        cpl_msg_error(__func__, "Wrong boundaries %d pix -> %d pix", 
                pix_min, pix_max) ;
        return NULL ;
    }

    /* Zero the rejected parts */
    spec_extract = cpl_vector_duplicate(spec) ;
    pspec = cpl_vector_get_data(spec_extract) ;
    for (i=0 ; i<cpl_vector_get_size(spec) ; i++) {
        if (i+1 < pix_min) pspec[i] = 0.0 ;
        if (i+1 > pix_max) pspec[i] = 0.0 ;
    }

    /* Clean the spectrum from the low frequency signal if requested */
    if (clean_spec) {
        cpl_msg_info(__func__, "Low Frequency signal removal") ;
        cpl_msg_indent_more() ;
        /* Subrtract the low frequency part */
        if ((filtered=cpl_vector_filter_median_create(spec_extract, 
                        filt_size))==NULL){
            cpl_msg_error(__func__, "Cannot filter the spectrum") ;
            cpl_msg_indent_less() ;
            return NULL ;
        }
        spec_clean = cpl_vector_duplicate(spec_extract) ;
        cpl_vector_subtract(spec_clean, filtered) ;
        cpl_vector_delete(filtered) ;
        cpl_msg_indent_less() ;
    } else {
        spec_clean = cpl_vector_duplicate(spec_extract) ;
    }
    cpl_vector_delete(spec_extract) ;
    
    /* Extract the interesting part of the catalog */
    cpl_msg_info(__func__, "Sub-Catalog extraction") ;
    cpl_msg_indent_more() ;
    catal_loc = irplib_wlxcorr_cat_extract(catalog_spec,
            wl_start-wl_err_wl, wl_stop+wl_err_wl) ;
    if (catal_loc == NULL) {
        cpl_msg_error(__func__, "Cannot extract sub-catalog [%g %g]",
                wl_start, wl_stop) ;
        /* Should not raise an error if there is no line */
        cpl_error_reset() ;
        /********/
        cpl_msg_indent_less() ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_indent_less() ;
  
    /* Zero the rejected parts */
    catal_loc_wl = cpl_bivector_get_x_data(catal_loc) ;
    catal_loc_intens = cpl_bivector_get_y_data(catal_loc) ;
    for (i=0 ; i<cpl_bivector_get_size(catal_loc) ; i++) {
        if (wl_min > 0) {
            if (catal_loc_wl[i] < wl_min) catal_loc_intens[i] = 0.0 ;
        }
        if (wl_max > 0) {
            if (catal_loc_wl[i] > wl_max) catal_loc_intens[i] = 0.0 ;
        }
    }

    /* Plot inputs */
    if (display) {
        irplib_wlxcorr_catalog_plot(catal_loc, wl_start, wl_stop) ;
        cpl_plot_vector(
    "set grid;set xlabel 'Position (Pixel)';set ylabel 'Intensity (ADU/sec)';",
    "t 'Extracted spectrum' w lines", "", spec_clean);
    }

    /* First solution */
    cpl_msg_info(__func__, "Find the best linear solution") ;
    cpl_msg_indent_more() ;
    wl_errors = cpl_vector_new(2) ;
    cpl_vector_fill(wl_errors, wl_err_wl) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", pow(nsamples, 2)) ;
    if ((xcdisp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, 1, phdisp, 
                    wl_errors, nsamples, slit_width, fwhm, &xc, 
                    NULL, &xcorrs)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate") ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        if (xcorrs != NULL) cpl_vector_delete(xcorrs) ;
        cpl_error_reset() ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
   
    /* Plot the first correlation values */
    if (display) {
        cpl_plot_vector("set grid;", 
                "t 'Correlation values for the linear solution search' w lines",
                "", xcorrs) ;
    }
    if (xcorrs != NULL) cpl_vector_delete(xcorrs) ;
    cpl_vector_delete(wl_errors) ;

    /* Higher degree --> less samples  + Lower error --> best precision */
    cpl_msg_info(__func__, "Find the best degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    wl_errors = cpl_vector_new(degree+1) ;
    cpl_vector_fill(wl_errors, wl_err_wl/wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/wlerr_factor) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, NULL, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;

    /* Lower error --> best precision */
    cpl_msg_info(__func__, "Refine the degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    cpl_vector_divide_scalar(wl_errors, wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/pow(wlerr_factor,2)) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, NULL, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;

    /* Lower error --> best precision */
    cpl_msg_info(__func__, "Refine the degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    cpl_vector_divide_scalar(wl_errors, wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/pow(wlerr_factor,3)) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, &wlres_loc, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;
    cpl_vector_delete(wl_errors) ;

    /* Plot results table */
    if (display) {
        /* Create the spc_table  */
        if ((spc_table = irplib_wlxcorr_gen_spc_table(spec_clean, catal_loc,
                        slit_width, fwhm, phdisp, xcdisp)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot generate infos table") ;
            cpl_polynomial_delete(xcdisp) ;
            cpl_bivector_delete(catal_loc) ;
            cpl_vector_delete(spec_clean) ;
            cpl_table_delete(wlres_loc) ;
            return NULL ;
        }
        /* Plot */
        irplib_wlxcorr_plot_spc_table(spc_table, "XC", fplotted, lplotted) ;
        cpl_table_delete(spc_table) ;
    }
    cpl_bivector_delete(catal_loc) ;
    
    /* Test Cross correlation factor */
    if (xc < xc_limit) {
        cpl_msg_warning(__func__, 
                "Cross-correlation too low (%g). Use estimate", xc) ;
        *xcorr = -1.0 ;
        cpl_polynomial_delete(xcdisp) ;
        xcdisp = cpl_polynomial_duplicate(phdisp) ;
    } else {
        cpl_msg_info(__func__, "Cross-correlation: %g", xc) ;
        *xcorr = xc ;
    }

    /* If PPM has been requested */
    if (use_ppm) {
        /* Threshold the clean spectrum */
        cpl_msg_info(__func__, "Threshold the spectrum") ;
        stdev = cpl_vector_get_stdev(spec_clean) ;
        median = cpl_vector_get_median_const(spec_clean) ;
        threshold = median + 0.2 * stdev ;
        pspec = cpl_vector_get_data(spec_clean) ;
        for (i=0 ; i<cpl_vector_get_size(spec_clean) ; i++) {
            if (pspec[i] < threshold) pspec[i] = 0.0 ;
        }
 
        /* Wavelength calibration with the Point-Pattern matching */
        cpl_msg_info(__func__, "Point Pattern Matching") ;
        if ((wl_poly_ppm = irplib_ppm_engine(spec_clean, catalog_spec, xcdisp,
                        5, 5, 0.2, degree, display, NULL)) == NULL) {
            cpl_msg_warning(__func__, "Cannot calibrate") ;
            wl_poly_ppm = xcdisp ;
        } else cpl_polynomial_delete(xcdisp) ;
    } else {
        wl_poly_ppm = xcdisp ;
    }
    cpl_vector_delete(spec_clean) ;

    /* Return the wlres table if requested */
    if (wlres != NULL) *wlres = wlres_loc ;
    else cpl_table_delete(wlres_loc) ;

    return wl_poly_ppm ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration map generation for 1 chip
  @param    wl_tab      table with the wavelengths solution
  @return   the wavelength map image or NULL in error case

  A 2D wavelength map is generated from the wavelength calibration table.
  For each y position, the 3 closest wavelength calibrations (with XC>0) are 
  used to interpolate (2nd degree polynomial) the current y position wavelength.

  If Only 2 solutions are available, a linear interpolation is done.

  If Only 1 is available, it is used as it is.

  If no solution with XC>0 is found, the closest is used.

  If the table does not contain the XC or YPOS information, the first row 
  is used.

  The returned image has to be deallocated with cpl_image_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_wlcalib_gen_wlmap_one_chip(const cpl_table * wl_tab)
{
    cpl_image           *   wl_map ;
    float               *   pwl_map ;
    int                     nx, ny ;
    cpl_polynomial      *   wl_poly1 ;
    cpl_polynomial      *   wl_poly2 ;
    cpl_table           *   wl_tab_loc ;
    cpl_table           *   wl_tab_valid ;
    int                     nb_valid, row1, row2 ;
    double                  y1, y2, wl1, wl2 ;
    int                     i, j ;

    /* Test entries */
    if (wl_tab == NULL) return NULL ;

    /* Initialise */
    nx = 1024 ;
    ny = 512 ;

    /* Create the image */
    wl_map = cpl_image_new(nx, ny, CPL_TYPE_FLOAT) ;
    pwl_map = cpl_image_get_data_float(wl_map) ;

    /* No Y pos info stored */
    if (!cpl_table_has_column(wl_tab, CRIRES_COL_WL_XCORR) ||
            !cpl_table_has_column(wl_tab, CRIRES_COL_WL_POSY)) {
        cpl_msg_info(__func__, "No YPOS information available - uniform Map") ;
        /* Get the first polynomial (there should be only one) */
        wl_poly1 = crires_wlcalib_gen_wlpoly(wl_tab, 0) ;
        
        /* Loop on the y positions */
        for (j=0 ; j<ny ; j++) {
            if (wl_poly1 == NULL) {
                for (i=0 ; i<nx ; i++)  pwl_map[i + j*nx] = 0.0 ;
            } else {
                /* Fill the image */
                for (i=0 ; i<nx ; i++)  pwl_map[i + j*nx] = 
                    cpl_polynomial_eval_1d(wl_poly1, (double)i+1, NULL) ;
            }
        }
        if (wl_poly1 != NULL) cpl_polynomial_delete(wl_poly1) ;
    } else {
        wl_tab_loc = cpl_table_duplicate(wl_tab) ;
       
        /* Get the solutions with xcorr > 0 */
        cpl_table_and_selected_double(wl_tab_loc,
                CRIRES_COL_WL_XCORR, CPL_GREATER_THAN, 0.0) ;
        nb_valid = (int)cpl_table_count_selected(wl_tab_loc) ;
        
        /* Only 0 or 1 solution */
        if (nb_valid <= 1) {
            if (nb_valid <= 0) {
                cpl_msg_info(__func__, 
                        "No computed solution available - uniform Map") ;
                /* No valid solution - Take any solution */
                wl_poly1 = crires_wlcalib_gen_wlpoly(wl_tab_loc, 0) ;
            } else if (nb_valid == 1) {
                cpl_msg_info(__func__, 
                        "One computed solution available - uniform Map") ;
                /* Use the one valid solution */
                wl_tab_valid =cpl_table_extract_selected(wl_tab_loc) ;
                wl_poly1 = crires_wlcalib_gen_wlpoly(wl_tab_valid, 0) ;
                cpl_table_delete(wl_tab_valid) ;
            }

            /* Loop on the y positions */
            for (j=0 ; j<ny ; j++) {
                if (wl_poly1 == NULL) {
                    for (i=0 ; i<nx ; i++)  pwl_map[i + j*nx] = 0.0 ;
                } else {
                    /* Fill the image */
                    for (i=0 ; i<nx ; i++)  pwl_map[i + j*nx] =
                        cpl_polynomial_eval_1d(wl_poly1, (double)i+1, NULL) ;
                }
            }
            if (wl_poly1 != NULL) cpl_polynomial_delete(wl_poly1) ;

        } else {
            /* Find the 2 closest solutions and interpolate */
            wl_tab_valid = cpl_table_extract_selected(wl_tab_loc) ;
            cpl_msg_info(__func__, "%d computed solution available", 
                    (int)cpl_table_get_nrow(wl_tab_valid)) ;
            /* Loop on the y positions */
            for (j=0 ; j<ny ; j++) {

                /* Get the 2 closest valid sol. from current Y pos. */
                if (crires_wlcalib_get_2_closest(wl_tab_valid, 
                            j+1, &row1, &row2) == -1) {
                    cpl_table_delete(wl_tab_loc) ;
                    cpl_image_delete(wl_map) ;
                    cpl_table_delete(wl_tab_valid) ;
                    return NULL ;
                }

                /* Get the polynomials */
                wl_poly1 = crires_wlcalib_gen_wlpoly(wl_tab_valid, row1) ;
                wl_poly2 = crires_wlcalib_gen_wlpoly(wl_tab_valid, row2) ;
                    
                if (wl_poly1 == NULL || wl_poly2 == NULL) {
                    for (i=0 ; i<nx ; i++)  pwl_map[i + j*nx] = 0.0 ;
                } else {
                    y1 = cpl_table_get(wl_tab_valid, 
                                CRIRES_COL_WL_POSY, row1, NULL) ;
                    y2 = cpl_table_get(wl_tab_valid, 
                                CRIRES_COL_WL_POSY, row2, NULL) ;

                    /* Fill the image by linear interpolation */
                    for (i=0 ; i<nx ; i++) {
                        wl1=cpl_polynomial_eval_1d(wl_poly1,(double)i+1,NULL) ;
                        wl2=cpl_polynomial_eval_1d(wl_poly2,(double)i+1,NULL) ;
                        pwl_map[i+j*nx] = ((wl2-wl1)/(y2-y1))*(j+1-y2)+wl2 ;
                    }
                }
                if (wl_poly1 != NULL) cpl_polynomial_delete(wl_poly1) ;
                if (wl_poly2 != NULL) cpl_polynomial_delete(wl_poly2) ;
            }
            cpl_table_delete(wl_tab_valid) ;
        }
        cpl_table_delete(wl_tab_loc) ;
    }
    return wl_map ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration map generation
  @param    wl_tab      table with the wavelengths solutions
  @return   the wavelength map images or NULL in error case

  A 2D wavelength map is generated from the wavelength calibration table.

  The returned image has to be deallocated with cpl_imagelist_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_wlcalib_gen_wlmap(const cpl_table ** wl_tab)
{
    cpl_imagelist       *   wl_map_list ;
    cpl_image           *   wl_map ;
    int                     i ;

    /* Test entries */
    if (wl_tab == NULL) return NULL ;
    for (i=0; i<CRIRES_NB_DETECTORS ; i++) {
        if (wl_tab[i] == NULL) return NULL ;
    }

    /* Create the output image list */
    wl_map_list = cpl_imagelist_new() ;

    /* Loop over the detectors */
    for (i=0; i<CRIRES_NB_DETECTORS ; i++) {
        /* Compute the map */
        wl_map = crires_wlcalib_gen_wlmap_one_chip(wl_tab[i]) ;
        if (wl_map == NULL) {
            cpl_imagelist_delete(wl_map_list) ;
            return NULL ;
        }

        /* Put the image in the list */
        cpl_imagelist_set(wl_map_list, wl_map, i) ;
    }
    return wl_map_list ;
}
/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration table solution (one row) generation
  @param    wl_poly     Polynomial relation (wl=P(x)) x from 1 to 1024
  @param    yposition   the y position in pixels of the solution (-1 for all) 
  @param    xcorr       The Xcorr computed to get the solution (-1.0 if none)
  @return   the wavelength solution table or NULL

  The returned table has to be deallocated with cpl_table_delete().
  It contains 
    CRIRES_COL_WL_COEFFA
    CRIRES_COL_WL_COEFFB
    CRIRES_COL_WL_COEFFC
    CRIRES_COL_WL_COEFFD
    CRIRES_COL_WL_COEFFE
    CRIRES_COL_WL_POSY
    CRIRES_COL_WL_XCORR

 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_wlcalib_gen_wltab_one(
        const cpl_polynomial    *   wl_poly, 
        int                         yposition, 
        double                      xcorr) 
{
    cpl_table       *   result ;
    cpl_size            power ;

    /* Check Entries */
    if (wl_poly == NULL) return NULL ;

    /* Create table */
    result = cpl_table_new(1);
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFA, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFB, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFC, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFD, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFE, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_POSY,   CPL_TYPE_INT) ;
    cpl_table_new_column(result, CRIRES_COL_WL_XCORR,  CPL_TYPE_DOUBLE) ;

    power = 0 ;
    cpl_table_set_double(result, CRIRES_COL_WL_COEFFA, 0, 
            cpl_polynomial_get_coeff(wl_poly, &power)) ;
    power = 1 ;
    cpl_table_set_double(result, CRIRES_COL_WL_COEFFB, 0, 
            cpl_polynomial_get_coeff(wl_poly, &power)) ;
    power = 2 ;
    cpl_table_set_double(result, CRIRES_COL_WL_COEFFC, 0, 
            cpl_polynomial_get_coeff(wl_poly, &power)) ;
    power = 3 ;
    cpl_table_set_double(result, CRIRES_COL_WL_COEFFD, 0, 
            cpl_polynomial_get_coeff(wl_poly, &power)) ;
    power = 4 ;
    cpl_table_set_double(result, CRIRES_COL_WL_COEFFE, 0, 
            cpl_polynomial_get_coeff(wl_poly, &power)) ;
    cpl_table_set_int(result, CRIRES_COL_WL_POSY, 0, yposition) ;
    cpl_table_set_double(result, CRIRES_COL_WL_XCORR, 0, xcorr) ;

    return result ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration table solution (several rows) generation
  @param    wl_polys    Polynomial relations array (wl=P(x)) x from 1 to 1024
  @param    ypositions  The y positions in pixels 
  @param    xcorrs      The Xcorrs array
  @return   the wavelength solution table or NULL

  The returned table has to be deallocated with cpl_table_delete().
  It contains 
    CRIRES_COL_WL_COEFFA
    CRIRES_COL_WL_COEFFB
    CRIRES_COL_WL_COEFFC
    CRIRES_COL_WL_COEFFD
    CRIRES_COL_WL_COEFFE
    CRIRES_COL_WL_POSY
    CRIRES_COL_WL_XCORR

 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_wlcalib_gen_wltab_several(
        const cpl_polynomial    **  wl_polys, 
        const cpl_vector        *   ypositions, 
        const double            *   xcorrs) 
{
    cpl_table       *   result ;
    cpl_size            power ;
    int                 nb_sol ;
    int                 i ;

    /* Check Entries */
    if (ypositions == NULL) return NULL ;
    nb_sol = cpl_vector_get_size(ypositions) ;
    if (wl_polys == NULL) return NULL ;
    if (xcorrs == NULL) return NULL ;
    for (i=0 ; i<nb_sol ; i++) {
        if (wl_polys[i] == NULL) return NULL ;
    }

    /* Create table */
    result = cpl_table_new(nb_sol);
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFA, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFB, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFC, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFD, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_COEFFE, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(result, CRIRES_COL_WL_POSY,   CPL_TYPE_INT) ;
    cpl_table_new_column(result, CRIRES_COL_WL_XCORR,  CPL_TYPE_DOUBLE) ;

    /* Loop on the solutions */
    for (i=0 ; i<nb_sol ; i++) {
        power = 0 ;
        cpl_table_set_double(result, CRIRES_COL_WL_COEFFA, i, 
                cpl_polynomial_get_coeff(wl_polys[i], &power)) ;
        power = 1 ;
        cpl_table_set_double(result, CRIRES_COL_WL_COEFFB, i, 
                cpl_polynomial_get_coeff(wl_polys[i], &power)) ;
        power = 2 ;
        cpl_table_set_double(result, CRIRES_COL_WL_COEFFC, i, 
                cpl_polynomial_get_coeff(wl_polys[i], &power)) ;
        power = 3 ;
        cpl_table_set_double(result, CRIRES_COL_WL_COEFFD, i, 
                cpl_polynomial_get_coeff(wl_polys[i], &power)) ;
        power = 4 ;
        cpl_table_set_double(result, CRIRES_COL_WL_COEFFE, i, 
                cpl_polynomial_get_coeff(wl_polys[i], &power)) ;
        cpl_table_set_int(result, CRIRES_COL_WL_POSY, i, 
                cpl_vector_get(ypositions, i)) ;
        cpl_table_set_double(result, CRIRES_COL_WL_XCORR, i, xcorrs[i]) ;
    }

    return result ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Automatic detection of the ypos vector
  @param    fname   Image with the pinhole 
  @param    period
  @param    the chip number
  @return   the vector with the positions
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_wlcalib_detect_wave_ypos(
        const char          *   fname,
        crires_illum_period     period,
        int                     chip)
{
    cpl_vector      *   positions ;
    cpl_image       *   in ;
    cpl_image       *   in_tmp ;
    cpl_image       *   collapsed ;
    cpl_image       *   filtered ;
    cpl_matrix      *   kernel ;
    cpl_size            xpos, ypos, nx, ny, nsamples ;
    int                 starty, stopy ;
    float           *   pima ;
    double              max ;
    int                 hsize, hsize2, nb_detected, start_ind, stop_ind ;
    cpl_vector      *   vec ;
    int                 i, j ;

    /* Test entries */
    if (fname == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    hsize = 20 ;
    hsize2 = 10 ;
    nb_detected = 6 ;

    /* Load the frame */
    if ((in = crires_load_chip(fname, period, chip, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the chip %d", chip) ;
        return NULL ;
    }
    nx = cpl_image_get_size_x(in) ;
    ny = cpl_image_get_size_y(in) ;

    /* Collapse the image along the X direction */
    collapsed = cpl_image_collapse_create(in, 1) ;
    
    /* Get the Max Y position */
    cpl_image_get_maxpos(collapsed, &xpos, &ypos) ;
    cpl_image_delete(collapsed) ;

    /* printf("Ypos max : %d\n", (int)ypos) ;  */

    /* Erase the zone around ypos in the full image */
    in_tmp = cpl_image_duplicate(in) ;
    pima = cpl_image_get_data_float(in_tmp);
    starty = ypos - hsize ;
    stopy = ypos + hsize ;
    if (starty < 1) starty = 1 ;
    if (stopy > ny) stopy = ny ;
    for (j=starty-1 ; j<=stopy-1 ; j++) {
        for (i=0 ; i<nx ; i++) {
            pima[i+j*nx] = 0.0 ;
        }
    }

    /* Collapse the image along the Y direction */
    collapsed = cpl_image_collapse_create(in_tmp, 0) ;
    cpl_image_delete(in_tmp) ;

    /* Lowpass */
    filtered = cpl_image_duplicate(collapsed) ;
    kernel = cpl_matrix_new(1, 3) ;
    cpl_matrix_fill(kernel, 1.0) ;
    cpl_image_filter(filtered, collapsed, kernel, CPL_FILTER_LINEAR,
            CPL_BORDER_FILTER);
    cpl_matrix_delete(kernel) ;
    cpl_image_delete(collapsed) ;
    
    /* vec = cpl_vector_new_from_image_row(filtered, 1); */
    /* cpl_plot_vector(NULL, NULL, NULL, vec) ; */
    /* cpl_vector_delete(vec) ; */

    /* Get the Max X position */
    cpl_image_get_maxpos_window(filtered, hsize, 1, ny-hsize, 1, &xpos, &ypos);
    cpl_image_delete(filtered) ;

    /* printf("Xpos max : %d\n", (int)xpos) ;  */
    collapsed = cpl_image_collapse_window_create(in, xpos-hsize, 1, 
            xpos+hsize, ny, 1) ;
    cpl_image_delete(in) ;
    vec = cpl_vector_new_from_image_column(collapsed, 1);
    cpl_image_delete(collapsed) ;

    /* cpl_plot_vector(NULL, NULL, NULL, vec) ; */
    
    positions = cpl_vector_new(nb_detected);
    cpl_vector_fill(positions, -1.0) ;
    nsamples = cpl_vector_get_size(vec) ;
    for (i=0 ; i<nb_detected ; i++) {
        
        /* Stop if everything is 0 */
        if ((max = cpl_vector_get_max(vec)) <= 0.0) break ;

        /* Find the strongest line */
        for (j=0 ; j<nsamples ; j++) {
            if (cpl_vector_get(vec, j) == max) break ;
        }
        cpl_vector_set(positions, i, (double)j+1) ;
        /* Zero the strongest line */
        if (j-hsize2 < 0) start_ind = 0 ;
        else start_ind = j-hsize2 ;
        if (j+hsize2 > nsamples-1) stop_ind = nsamples-1 ;
        else stop_ind = j+hsize2 ;
        for (j=start_ind ; j<=stop_ind ; j++) cpl_vector_set(vec, j, 0.0) ;

    }
    cpl_vector_delete(vec) ;
    cpl_vector_sort(positions, CPL_SORT_ASCENDING) ;
    /* cpl_vector_dump(positions, stdout) ; */

    return positions ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the best XC stored in the passed wave_cal table
  @param    wl_tab      The wave_cal table
  @return   the best XC or -1
 */
/*----------------------------------------------------------------------------*/
double crires_wlcalib_get_better_xc(const cpl_table * wl_tab)
{
    int     nrows ;
    double  best_xc, curr_xc ;
    int     i ;
   
    /* Test entries */
    if (wl_tab == NULL) return -1.0 ;
    if (!cpl_table_has_column(wl_tab, CRIRES_COL_WL_XCORR)) return -1.0 ;

    /* Initialise */
    best_xc = -1.0 ;

    /* Get the number of rows */
    nrows = cpl_table_get_nrow(wl_tab) ;

    /* Loop over the solutions */
    for (i=0 ; i<nrows ; i++) {
        curr_xc = cpl_table_get(wl_tab, CRIRES_COL_WL_XCORR, 0, NULL) ;
        if (curr_xc > best_xc) best_xc = curr_xc ;
    }
    return best_xc ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration polynomial generation
  @param    wl_tab      table with the wavelengths solutions
  @param    row         the table row (0 -> ... )
  @return   the wavelength polynomial or NULL in error case

  The returned polynomial has to be deallocated with cpl_polynomial_delete().
 */
/*----------------------------------------------------------------------------*/
static cpl_polynomial * crires_wlcalib_gen_wlpoly(
        const cpl_table     *   wl_tab, 
        int                     row)
{
    cpl_polynomial      *   wl_poly ;
    double                  a, b, c, d, e ;
    int                     nrows ;
    cpl_size                power ;

    /* Test entries */
    if (wl_tab == NULL) return NULL ;

    /* Initialise */
    nrows = cpl_table_get_nrow(wl_tab) ;
    if (row >= nrows) return NULL ;

    /* Get the coefficients */
    a = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFA, row, NULL) ;
    b = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFB, row, NULL) ;
    c = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFC, row, NULL) ;
    d = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFD, row, NULL) ;
    e = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFE, row, NULL) ;

    /* Create the polynomial */
    wl_poly = cpl_polynomial_new(1) ;
    power = 0 ; cpl_polynomial_set_coeff(wl_poly, &power, a) ;
    power = 1 ; cpl_polynomial_set_coeff(wl_poly, &power, b) ;
    power = 2 ; cpl_polynomial_set_coeff(wl_poly, &power, c) ;
    power = 3 ; cpl_polynomial_set_coeff(wl_poly, &power, d) ;
    power = 4 ; cpl_polynomial_set_coeff(wl_poly, &power, e) ;
    
    return wl_poly ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the closest rows from a given Y position
  @param    tab     the Table with the wavelength solutions
  @param    ypos    the Y position (1 -> )
  @param    row1    The computed row
  @param    row2    The computed row
  @return   0 if 0k, -1 otherwise.
 */
/*----------------------------------------------------------------------------*/
static int crires_wlcalib_get_2_closest(
        const cpl_table *   tab, 
        int                 ypos, 
        int             *   row1,
        int             *   row2)
{
    double  min_dist, curr_dist ;
    int     i ;

    /* Test entries */
    if (tab == NULL) return -1 ;
    if (row1 == NULL || row2 == NULL) return -1 ;

    /* Initialise */
    *row1 = *row2 = -1 ;
    min_dist = 500 ;
    
    /* Find the closest solution */
    for (i=0 ; i<cpl_table_get_nrow(tab) ; i++) {
        curr_dist=fabs(cpl_table_get(tab,CRIRES_COL_WL_POSY,i,NULL)-ypos);
        if (curr_dist < min_dist) {
            min_dist = curr_dist ;
            *row1 = i ;
        }
    }

    /* Find the second closest */
    min_dist = 500 ;
    for (i=0 ; i<cpl_table_get_nrow(tab) ; i++) {
        if (i != *row1) {
            curr_dist=fabs(cpl_table_get(tab,CRIRES_COL_WL_POSY,i,NULL)-ypos);
            if (curr_dist < min_dist) {
                min_dist = curr_dist ;
                *row2 = i ;
            }
        }
    }
    if (*row1 < 0 || *row2 < 0) return -1 ;
    return 0 ;
}

/**@}*/
