/* $Id: crires_model_refine.c,v 1.55 2012-01-16 10:09:55 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-01-16 10:09:55 $
 * $Revision: 1.55 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

#include "irplib_spectrum.h"

#include "crires_model_refining.h"
#include "crires_model_kernel.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_model_refine"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_model_refine_save(cpl_table *, const cpl_parameterlist *, 
        cpl_frameset *) ;

static char crires_model_refine_description[] =
"crires_model_refine -- Model refining recipe\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "CRIRES_MODEL_REFINE_RAW" or\n"
"THAR-file.fits "CRIRES_CALPRO_THAR_CAT" or\n"
"bpm-file.fits "CRIRES_CALPRO_BPM" or\n"
"flat-file.fits "CRIRES_CALPRO_FLAT" or\n"
"detlin-file.fits "CRIRES_CALPRO_COEFFS_CUBE" or\n"
"config-file.fits "CRIRES_CALPRO_MODEL_REFINE_CONF".\n" ;

CRIRES_RECIPE_DEFINE(crires_model_refine,
		     CRIRES_PARAM_DISPLAY        |
		     CRIRES_PARAM_FWHM           |
		     CRIRES_PARAM_SIGMA          |
		     CRIRES_PARAM_SEARCH_BOX     |
		     CRIRES_PARAM_MIN_MATCH,      
		     "Model Refining recipe",
		     crires_model_refine_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             display ;
    int             search_box_sz ;
    int             min_matches_nb ;
    int             fwhm ;
    double          sigma ;
    /* Outputs */
    crires_illum_period period ;
    double              penc_co1 ;
    double              genc_co1 ;
    double              sg ;
    double              fk ;
    double              cam_dis ;
    double              slity ;
} crires_model_refine_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parlist     the parameters list
  @param    frameset   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_model_refine(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    cpl_frameset        *   rawframes ;
    const char          *   config ;
    const char          *   thar ;
    const char          *   bpm ;
    const char          *   flat ;
    const char          *   detlin ;
    const char          *   ref_fname=NULL ;
    cpl_frame           *   ref_frame ;
    cpl_vector          **  traces ;
    double              *   ptrace ;
    double              **  cust_lines=NULL;
    cpl_imagelist       *   ilist ;
    cpl_vector          *   extracted[CRIRES_NB_DETECTORS] ;
    cpl_vector          *   bright_lines[CRIRES_NB_DETECTORS] ;
    cpl_table           *   new_conf ;
    int                     nframes;
    int                     i, j, k;

    double y_pos_mm=-50;
    int coord_cnt=0;
    double ** temp_coord;
    coord msp_coord[400];
    int ph, cust_lines_flag;
    int nph=7;
    double y_pos_pix=0.0;
    FILE* scan_out;

    /* The Model is switched off */
    if (crires_model_off()) {
        return 0 ;
    }

    /* Initialise */
    rawframes = NULL ;
    crires_model_refine_config.penc_co1 = -1.0 ;
    crires_model_refine_config.genc_co1 = -1.0 ;
    crires_model_refine_config.sg = -1.0 ;
    crires_model_refine_config.fk = -1.0 ;
    crires_model_refine_config.cam_dis = -1.0 ;
    crires_model_refine_config.slity = -1.0 ;
    
    /* Retrieve input parameters */
    crires_model_refine_config.display = crires_parameterlist_get_int(parlist, 
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_model_refine_config.fwhm = crires_parameterlist_get_int(parlist, 
            RECIPE_STRING, CRIRES_PARAM_FWHM) ; 
    crires_model_refine_config.sigma = crires_parameterlist_get_double(parlist,
            RECIPE_STRING, CRIRES_PARAM_SIGMA) ;
    crires_model_refine_config.search_box_sz = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_SEARCH_BOX) ;
    crires_model_refine_config.min_matches_nb = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_MIN_MATCH) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, "crires_model_refine")) {
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Retrieve calibration data */
    config = crires_extract_filename(frameset, CRIRES_CALPRO_MODEL_REFINE_CONF);
    if (config == NULL) {
        cpl_msg_error(__func__, "Model configuration file is missing") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    thar = crires_extract_filename(frameset, CRIRES_CALPRO_THAR_CAT) ;
    if (thar == NULL) {
        cpl_msg_error(__func__, "Lines catalog file is missing") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    bpm         = crires_extract_filename(frameset, CRIRES_CALPRO_BPM) ;
    flat        = crires_extract_filename(frameset, CRIRES_CALPRO_FLAT) ;
    detlin      = crires_extract_filename(frameset, CRIRES_CALPRO_COEFFS_CUBE) ;

    /* Retrieve raw frames */
    if ((rawframes = crires_extract_frameset(frameset, 
                    CRIRES_MODEL_REFINE_RAW)) == NULL) {
        cpl_msg_error(__func__, "No raw frame in input") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    nframes = cpl_frameset_get_size(rawframes) ;

    /* Get the detector illumination period */
    crires_model_refine_config.period =
        crires_get_detector_illum_period(
            cpl_frame_get_filename(cpl_frameset_get_position(rawframes, 0))) ;
    if (crires_model_refine_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__,
                "Cannot determine the detector illumination period") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        cpl_frameset_delete(rawframes) ;
        return -1 ;
    } else {
        crires_display_detector_illum(crires_model_refine_config.period) ;
    }

    /* Check the number of input frames */
    if (nframes != 1 && nframes != 3) {
        cpl_msg_error(__func__, "1 or 3 raw frames expected in input") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }

    /* Loop on the input frames */
    for (k=0 ; k<nframes ; k++) {
        /* Get the reference frame */
        ref_frame = cpl_frameset_get_position(rawframes, k) ;
        ref_fname = cpl_frame_get_filename(ref_frame) ;

	if (crires_model_config_check(config,ref_fname)!=0) {
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
	  cpl_msg_error(__func__,
                "Invalid physical model configuration file") ;
        cpl_frameset_delete(rawframes) ;
	return -1;
	}

	cust_lines_flag=0;
        for (ph=1;ph<=nph;ph++) {
	  y_pos_pix=-(float)(ph);
	  cpl_msg_info(__func__,"Raw frame %d, fibre position %d",k+1,ph);
	  /* Get the shape of the spectrum at the detector y posn specified */
	  cpl_msg_info(__func__, "Get the shape of the spectra") ;
	  if ((traces = crires_model_profile(ref_fname,
					     config,
					     -1,
					     y_pos_pix)) != NULL) {
	    /* Clean the traces */
	    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
	      ptrace = cpl_vector_get_data(traces[i]) ;
	      for (j=0 ; j<cpl_vector_get_size(traces[i]) ; j++) {
		if (ptrace[j] == 0.0) 
		  ptrace[j] = cpl_vector_get_mean(traces[i]) ;
	      }
	    }

	    /* Plot the shapes if requested */
	    if (crires_model_refine_config.display>0 &&
		crires_model_refine_config.display <= CRIRES_NB_DETECTORS) {
	      cpl_plot_vector("set grid;set xlabel 'Position (pixels)';set ylabel 'Position (pixels)';",
				 "t 'Spectrum shape' w lines", "", traces[crires_model_refine_config.display-1]);
	    }

	    /* Load the input images */
	    cpl_msg_info(__func__, "Load the data") ;
	    ilist = crires_load_file(ref_fname,
				     crires_model_refine_config.period, CPL_TYPE_FLOAT) ;
	    if (ilist == NULL) {
	      cpl_msg_error(__func__, "Cannot load the data") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
	      cpl_frameset_delete(rawframes) ;
	      for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
		cpl_vector_delete(traces[i]) ;
	      cpl_free(traces) ;
	      return -1 ;
	    }
	
	    /* Customise the line list */
	    if (cust_lines_flag==0) {
	      cpl_msg_info(__func__, "Customising the line list") ;
	      cpl_msg_indent_more() ;
	      if ((cust_lines=crires_model_refining_custom_lines(thar,
								 crires_model_refine_config.search_box_sz,
								 ref_fname, config, NULL))==NULL) {
		for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
		  cpl_vector_delete(traces[i]) ;
		cpl_free(traces) ;
		cpl_imagelist_delete(ilist) ;
		cpl_frameset_delete(rawframes) ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		cpl_msg_error(__func__, "Cannot customise the line list");
		return -1;
	      }
	      else {
		cust_lines_flag=1;
	      }
	    }
	    cpl_msg_indent_less() ;
	    
	    /* Update the cust_lines list to have the correct predicted 
	       detector x,y for the current pinhole*/
	    cust_lines=crires_model_refining_fibrefix(cust_lines,ref_fname,config,ph);

	    /* Loop on detectors */
	    cpl_msg_indent_more() ;
	    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
	      cpl_msg_info(__func__, "Reduce the chip number %d", i+1) ;
	      /* Extract the spectrum using the shape */
	      cpl_msg_info(__func__, "Extract spectrum") ;
	      cpl_msg_indent_more() ;
	      if ((extracted[i]=crires_model_refining_extract(cpl_imagelist_get(ilist, i),
							      crires_model_refine_config.period,
							      traces[i], bpm, flat, detlin, i+1)) == NULL) {
		cpl_msg_error(__func__, "Cannot extract spectrum");
		cpl_frameset_delete(rawframes) ;
		for (j=0 ;j<CRIRES_NB_DETECTORS; j++) 
		  cpl_vector_delete(traces[j]) ;
		cpl_free(traces) ;
		cpl_imagelist_delete(ilist) ;
		for (j=0 ; j<i ; j++) cpl_vector_delete(extracted[j]) ;
		if ((crires_model_free2Darray(cust_lines,6))!=0) {
		  cpl_msg_error(__func__, "Cannot free 2D array");
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		  return -1;
		}
		cpl_msg_indent_less() ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		return -1 ;
	      }
	      cpl_msg_indent_less() ;

	      /* Display if requested */
	      if (crires_model_refine_config.display == i+1) {
		cpl_plot_vector("set grid;set xlabel 'Position (pixels)';set ylabel 'Intensity (ADU)';",
				   "t 'Extracted spectrum' w lines", "", extracted[i]);
	      }

	      /* Detect the brightest lines */
	      cpl_msg_info(__func__, "Detect the brightest lines") ;
	      cpl_msg_indent_more() ;
	      if ((bright_lines[i] = irplib_spectrum_detect_peaks(extracted[i],
								  crires_model_refine_config.fwhm,
								  crires_model_refine_config.sigma,
								  crires_model_refine_config.display==i+1,
								  NULL, NULL)) == NULL) {
		cpl_msg_warning(__func__, "Cannot detect lines");
	      }
/* 		    for (pp=0;pp<cpl_vector_get_size(bright_lines[i]);pp++){ */
/* 		      printf("%lf %d %lf %d %lf %d %lf\n", */
/* 			     cpl_vector_get(bright_lines[i],pp), */
/* 			     (int)(cpl_vector_get(bright_lines[i],pp)-0.5)-1, */
/* 			     cpl_vector_get(extracted[i],(int)(cpl_vector_get(bright_lines[i],pp)-0.5)-1), */
/* 			     (int)(cpl_vector_get(bright_lines[i],pp)-0.5), */
/* 			     cpl_vector_get(extracted[i],(int)(cpl_vector_get(bright_lines[i],pp)-0.5)), */
/* 			     (int)(cpl_vector_get(bright_lines[i],pp)-0.5)+1, */
/* 			     cpl_vector_get(extracted[i],(int)(cpl_vector_get(bright_lines[i],pp)-0.5)+1)); */
/* 		      totflux=0.0; */
/* 		      moment=0.0; */
/* 		      for (qq=-3;qq<4;qq++) { */
/* 			flux=cpl_vector_get(extracted[i],(int)(cpl_vector_get(bright_lines[i],pp))+qq); */
/* 			if (flux>0.0) { */
/* 			  moment+=(float)(qq)*flux; */
/* 			  totflux+=flux; */
/* 			} */
/* 		      } */
/* 		      cpl_vector_set(bright_lines[i],pp,(float)((int)(cpl_vector_get(bright_lines[i],pp)))+(moment/totflux)); */
/*  		      printf("%d %lf %lf \n",pp, cpl_vector_get(bright_lines[i],pp),moment/totflux);  */
/* 		    } */
	      cpl_msg_indent_less() ;
	    }
	    cpl_imagelist_delete(ilist) ;
	    for (j=0 ; j<CRIRES_NB_DETECTORS; j++) 
	      cpl_vector_delete(extracted[j]) ;
    
	    /* Match with the catalog */
	    cpl_msg_info(__func__, "Match with the catalog") ;
	    cpl_msg_indent_more() ;
	    if ((temp_coord=crires_model_refining_match((const cpl_vector **)bright_lines,
							(const cpl_vector **)traces,
							cust_lines,
							crires_model_refine_config.search_box_sz,
							crires_model_refine_config.min_matches_nb,
							ref_fname)) == NULL) {
	      if (nph==1) {
		cpl_msg_error(__func__, "Cannot match %s with catalog",
			      ref_fname);
		cpl_frameset_delete(rawframes) ;
		for (j=0 ;j<CRIRES_NB_DETECTORS; j++) 
		  cpl_vector_delete(traces[j]) ;
		cpl_free(traces) ;
		for (j=0 ; j<CRIRES_NB_DETECTORS; j++) 
		  if (bright_lines[j] != NULL) 
		    cpl_vector_delete(bright_lines[j]) ;
		if ((crires_model_free2Darray(cust_lines,6))!=0) {
		  cpl_msg_error(__func__, "Cannot free 2D array");
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		  return -1;
		}
		cpl_msg_indent_less() ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		return -1 ;
	      } else { 
		cpl_msg_warning(__func__, "Cannot match lines for fibre %d on raw frame %d", ph, k);
	      }
	    }
	    cpl_msg_indent_less() ;
	    for (j=0 ; j<CRIRES_NB_DETECTORS; j++) 
	      if (bright_lines[j] != NULL) 
		cpl_vector_delete(bright_lines[j]) ;
	    if (temp_coord!=NULL) {
	      i=0;
	      while (temp_coord[0][i]!=0.0) {
		msp_coord[coord_cnt].wave=temp_coord[0][i];
		msp_coord[coord_cnt].x=temp_coord[1][i];
		msp_coord[coord_cnt].flux=temp_coord[2][i];
		msp_coord[coord_cnt].chip=(int)(temp_coord[3][i]);
		msp_coord[coord_cnt].mode=(int)(temp_coord[4][i]);
		msp_coord[coord_cnt].y=temp_coord[5][i];
		msp_coord[coord_cnt].penc=(int)(temp_coord[6][i]);
		msp_coord[coord_cnt].genc=(int)(temp_coord[7][i]);
		msp_coord[coord_cnt].slit_pos=ph;
		msp_coord[coord_cnt].counter=(int)(temp_coord[9][i]);
		msp_coord[coord_cnt].slitw=temp_coord[10][i];
		coord_cnt++;
		i++;
	      }
	      if ((crires_model_free2Darray(temp_coord,11))!=0) {
		cpl_msg_error(__func__, "Cannot free 2D array");
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
		return -1;
	      }
	      y_pos_mm=cpl_vector_get(traces[2],1024);
	    } /*end of match check*/
	    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
	      cpl_vector_delete(traces[i]) ;
	    cpl_free(traces);
	    cpl_msg_indent_less() ;
	  } /*end of trace check*/
	  else {
	    cpl_msg_warning(__func__, "Cannot get the spectrum trace") ;
	  }
    
        } /*end of ph loop*/
	if ((crires_model_free2Darray(cust_lines,6))!=0) {
	  cpl_msg_error(__func__, "Cannot free 2D array");
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
	  return -1;
	}
    } /*end of ds loop*/
    
    // TEMP SCAN OUTPUT
    if (cal_flag==1) {
      scan_out=fopen("coord_scan.dat","a+");
      for (i=0;i<coord_cnt;i++) {
        fprintf(scan_out,"%d %lf %lf %d %lf %d %lf %d %d %d %lf\n",i,
	        msp_coord[i].y,
	        msp_coord[i].x,
	        msp_coord[i].chip,
	        msp_coord[i].flux,
	        msp_coord[i].slit_pos,
	        msp_coord[i].wave*1000000.0,
	        -msp_coord[i].mode,
	        msp_coord[i].penc,
	        msp_coord[i].genc,
	        msp_coord[i].slitw);
      }
      fclose(scan_out);
    }

    cpl_msg_info(__func__, "slit position in mm: %lf",y_pos_mm);
    /* Compute the new configuration file */
    cpl_msg_info(__func__, "Compute the new configuration file") ;
    cpl_msg_indent_more() ;
    if ((new_conf = crires_model_refining_anneal(msp_coord,
                    coord_cnt,
                    y_pos_mm,
                    ref_fname,
                    config)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the new configuration file");
        cpl_msg_indent_less() ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1; 
    }
    cpl_msg_indent_less() ;
    cpl_frameset_delete(rawframes) ;

    /* Save the new configuration file */
    cpl_msg_info(__func__, "Save the new configuration file") ;
    if (crires_model_refine_save(new_conf, parlist, frameset) == -1) {
        cpl_msg_error(__func__, "Cannot save the product");
        cpl_table_delete(new_conf) ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }
    cpl_table_delete(new_conf) ;
   
    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}
    
    
/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table 
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int crires_model_refine_save(
        cpl_table               *   out_table,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    *   plist ;
    const char          *   recipe_name = "crires_model_refine" ;

    plist = cpl_propertylist_new();
    cpl_propertylist_append_string(plist, "INSTRUME", "CRIRES") ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_CATG,
            CRIRES_CALPRO_MODEL_CONFIG) ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_TYPE,
            CRIRES_PROTYPE_MOD_CONF) ;


    /* Save */
    if (cpl_dfs_save_table(set, NULL, parlist, set, NULL, out_table,
                NULL, recipe_name, plist, NULL, PACKAGE "/" PACKAGE_VERSION,
                "crires_model_refine.fits") != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the table") ;
        return -1 ;
    }

    /* Return */
    cpl_propertylist_delete(plist) ;
    return 0 ;
}

