/* 
 * This file is part of the crires Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <locale.h>
#include "crires_recipe.h"

#include "crires_extract.h"
#include "irplib_utils.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_extract"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_extract_save(const cpl_table **, const cpl_imagelist *, 
        const cpl_imagelist *, const cpl_parameterlist *, cpl_frameset *) ;

static char crires_util_extract_description[] = 
"This recipe accepts 1 parameter (optionally a 2nd one):\n"
"First parameter:   the combined image.\n" 
"                   (PRO TYPE = "CRIRES_PROTYPE_COMBINED")\n"
"Second parameter (optional):   the contribution map.\n" 
"                   (PRO TYPE = "CRIRES_PROTYPE_CONTRIB")\n"
"\n"
"This recipe produces 3 files:\n"
"First product:     the image of the profile of the spectrum\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_PROFILE")\n"
"Second product:    the table of the extracted spectrum in pixels\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_SPEC_PIX")\n"
"Third product:     the maps of the background\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_BGD_MAP")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_extract,
        CRIRES_PARAM_DISPLAY            |
        CRIRES_PARAM_EXTR_MODE          |
        CRIRES_PARAM_HOR_SIZE           |
        CRIRES_PARAM_SPEC_HSIZE         |
        CRIRES_PARAM_KAPPA              |
        CRIRES_PARAM_CLOSING_HSIZE      |
        CRIRES_PARAM_CLEAN_RATE         |
        CRIRES_PARAM_TOT_NDIT           |
        CRIRES_PARAM_Y_POS_CHIP1        |
        CRIRES_PARAM_Y_POS_CHIP2        |
        CRIRES_PARAM_Y_POS_CHIP3        |
        CRIRES_PARAM_Y_POS_CHIP4        |
        CRIRES_PARAM_Y_WIDTH            |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP1  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP2  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP3  |
        CRIRES_PARAM_Y_SPEC_ZONE_CHIP4  |
        CRIRES_PARAM_REJECT,
        "Spectrum extraction routine",
        crires_util_extract_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    /* mode : 1=spec ; 2=lines */
    int                 mode ;
    double              kappa ;         /* mode = 1 */
    int                 closing_hs ;    /* mode = 1 */
    double              clean_rate ;    /* mode = 1 */
    double              tot_ndit ;      /* mode = 1 */
    int                 box_hor_size ;  /* mode = 1 */
    int                 spec_hsize ;    /* mode = 1 */
    const char      *   y_spec_zone_c1 ;/* mode = 1 */
    const char      *   y_spec_zone_c2 ;/* mode = 1 */
    const char      *   y_spec_zone_c3 ;/* mode = 1 */
    const char      *   y_spec_zone_c4 ;/* mode = 1 */
    const char      *   y_pos_c1 ;      /* mode = 2 */
    const char      *   y_pos_c2 ;      /* mode = 2 */
    const char      *   y_pos_c3 ;      /* mode = 2 */
    const char      *   y_pos_c4 ;      /* mode = 2 */
    int                 y_width ;       /* mode = 2 */
    int                 rej_right ;     /* All modes */
    int                 rej_left ;      /* All modes */
    int                 display ;       /* All modes */
    /* Outputs */
    int                 win_mode ;
    crires_illum_period period ;
    int                 qc_specpos[CRIRES_NB_DETECTORS] ;
    int                 qc_specwrec[CRIRES_NB_DETECTORS] ;
    int                 qc_specwopt[CRIRES_NB_DETECTORS] ;
    double              qc_specoptmed[CRIRES_NB_DETECTORS] ;
    double              qc_s2nmed[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_comb_pix[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_comb_as[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_prof_pix[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_prof_as[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm_diff[CRIRES_NB_DETECTORS] ;
} crires_util_extract_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_extract(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist) 
{
    const char          *   sval ;
    const char          *   y_pos ;
    cpl_imagelist       *   imlist ;
    cpl_propertylist    *   plist ;
    cpl_imagelist       *   contrib ;
    cpl_image           *   contribution_map ;
    cpl_frame           *   fr ;
    cpl_imagelist       *   profile ;
    cpl_image           *   prof[CRIRES_NB_DETECTORS] ;
    cpl_imagelist       *   bgmaps ;
    cpl_image           *   bgmap[CRIRES_NB_DETECTORS] ;
    cpl_table           *   ext[CRIRES_NB_DETECTORS] ;
    cpl_vector          *   ext_vec ;
    cpl_vector          *   ypos_vec ;
    int                     ly, nexp, starty, stopy, extr_spec_starty,
                            extr_spec_stopy ;
    int                     i, j ;

    /* Needed for sscanf() */
    setlocale(LC_NUMERIC, "C");
    
    /* Initialise */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_util_extract_config.qc_specpos[i] = -1 ;
        crires_util_extract_config.qc_specwrec[i] = -1 ;
        crires_util_extract_config.qc_specwopt[i] = -1 ;
        crires_util_extract_config.qc_specoptmed[i] = -1.0 ;
        crires_util_extract_config.qc_s2nmed[i] = -1.0 ;
        crires_util_extract_config.qc_fwhm_comb_pix[i] = -1.0 ;
        crires_util_extract_config.qc_fwhm_comb_as[i] = -1.0 ;
        crires_util_extract_config.qc_fwhm_prof_pix[i] = -1.0 ;
        crires_util_extract_config.qc_fwhm_prof_as[i] = -1.0 ;
        crires_util_extract_config.qc_fwhm_diff[i] = -1.0 ;
        ext[i] = NULL ;
        prof[i] = NULL ;
        bgmap[i] = NULL ;
    }
    ly = -1 ;
    
    /* Retrieve input parameters */
    crires_util_extract_config.display = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_util_extract_config.mode = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_EXTR_MODE) ;
    crires_util_extract_config.box_hor_size = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_HOR_SIZE) ;
    crires_util_extract_config.spec_hsize = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_SPEC_HSIZE) ;
    crires_util_extract_config.kappa = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_KAPPA) ;
    crires_util_extract_config.closing_hs = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_CLOSING_HSIZE) ;
    crires_util_extract_config.clean_rate = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_CLEAN_RATE) ;
    crires_util_extract_config.tot_ndit = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_TOT_NDIT) ;
    crires_util_extract_config.y_pos_c1=crires_parameterlist_get_string(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP1) ;
    crires_util_extract_config.y_pos_c2=crires_parameterlist_get_string(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP2) ;
    crires_util_extract_config.y_pos_c3=crires_parameterlist_get_string(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP3) ;
    crires_util_extract_config.y_pos_c4=crires_parameterlist_get_string(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP4) ;
    crires_util_extract_config.y_width= crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_WIDTH) ;
    crires_util_extract_config.y_spec_zone_c1=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_SPEC_ZONE_CHIP1) ;
    crires_util_extract_config.y_spec_zone_c2=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_SPEC_ZONE_CHIP2) ;
    crires_util_extract_config.y_spec_zone_c3=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_SPEC_ZONE_CHIP3) ;
    crires_util_extract_config.y_spec_zone_c4=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_SPEC_ZONE_CHIP4) ;

    sval = crires_parameterlist_get_string(parlist, RECIPE_STRING,
            CRIRES_PARAM_REJECT) ;
    if (sscanf(sval, "%d,%d",
                    &crires_util_extract_config.rej_left,
                    &crires_util_extract_config.rej_right)!=2) {
        return -1 ;
    }
 
    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }
  
    /* Get the detector illumination period */
    fr = cpl_frameset_get_position(frameset, 0);
    crires_util_extract_config.period =
        crires_get_detector_illum_period(cpl_frame_get_filename(fr)) ;
    if (crires_util_extract_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__, 
                "Cannot determine the detector illumination period") ;
        return -1 ;
    }
 
    /* Windowing mode ? */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(fr), 0)) == NULL) 
        return -1 ;
    sval = crires_pfits_get_ncorrs(plist) ;
    if (!strcmp(sval, "FowlerNsampGRstWin")) {
        crires_util_extract_config.period = CRIRES_ILLUM_FULL_DETECTOR ;
        crires_util_extract_config.win_mode = 1 ;
    } else {
        crires_util_extract_config.win_mode = 0 ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Display the Detector illumination */
    crires_display_detector_illum(crires_util_extract_config.period) ;

    /* Load the first file in an image list */
    cpl_msg_info(__func__, "Input image loading") ;
    imlist = crires_load_file(cpl_frame_get_filename(fr), 
            crires_util_extract_config.period, CPL_TYPE_FLOAT) ;
  
    /* Load the optional contribution map */
    if (cpl_frameset_get_size(frameset) >= 2) {
        fr = cpl_frameset_get_position(frameset, 1);
        cpl_msg_info(__func__, "Contribution map specified: %s",
                cpl_frame_get_filename(fr)) ;
        contrib = crires_load_file(cpl_frame_get_filename(fr), 
                crires_util_extract_config.period, CPL_TYPE_INT) ;
    } else contrib = NULL ;
   
    if (crires_util_extract_config.mode == 1) {
        /* Get the total ndit */
        fr = cpl_frameset_get_position(frameset, 0);
        if (crires_util_extract_config.tot_ndit < 0) {
            crires_util_extract_config.tot_ndit = 
                crires_get_totndit(cpl_frame_get_filename(fr)) ;
            if (crires_util_extract_config.tot_ndit < 0) {
                cpl_msg_error(__func__, 
                        "Cannot get the total number of NDIT*DIT") ;
                cpl_imagelist_delete(imlist) ;
                if (contrib != NULL) cpl_imagelist_delete(contrib) ;
                return -1 ;
            }
        }
        /* Get the number of exposures used for the frames combination */
        plist=cpl_propertylist_load(cpl_frame_get_filename(fr), 0) ;
        nexp = crires_pfits_get_datancom(plist) ;
        cpl_propertylist_delete(plist) ;
        if (cpl_error_get_code() != CPL_ERROR_NONE) {
            cpl_msg_error(__func__,"No PRO DATANCOM information in the header");
            cpl_imagelist_delete(imlist) ;
            if (contrib != NULL) cpl_imagelist_delete(contrib) ;
            return -1 ;
        }
        crires_util_extract_config.tot_ndit *= nexp ;
    }
    
    /* Extract chips */
    cpl_msg_info(__func__, "Spectrum extraction") ;
    cpl_msg_indent_more() ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Skip some detectors in windowing mode */
        if ((i==0 || i==CRIRES_NB_DETECTORS-1) && 
                (crires_util_extract_config.win_mode == 1)) {
            continue ;
        }
        cpl_msg_info(__func__, "Spectrum extraction from chip number %d", i+1) ;
        cpl_msg_indent_more() ;
        if (crires_util_extract_config.mode == 1) {
            /* Where is the spectrum extraction zone ? */
            y_pos = "" ;
            if (i+1 == 1) y_pos = crires_util_extract_config.y_spec_zone_c1 ;
            if (i+1 == 2) y_pos = crires_util_extract_config.y_spec_zone_c2 ;
            if (i+1 == 3) y_pos = crires_util_extract_config.y_spec_zone_c3 ;
            if (i+1 == 4) y_pos = crires_util_extract_config.y_spec_zone_c4 ;
            if (sscanf(y_pos,"%d,%d", &extr_spec_starty, &extr_spec_stopy)!=2) {
                cpl_msg_warning(__func__, "Wrong Spectral Zone specified: %s",
                        y_pos) ;
                extr_spec_starty = extr_spec_stopy = -1 ;
            }

            /* Extract spectrum */
            if (contrib != NULL) 
                contribution_map = cpl_imagelist_get(contrib, i) ;
            else
                contribution_map = NULL ;
            ext[i] = crires_extract_spectrum(cpl_imagelist_get(imlist, i),
                    contribution_map,
                    crires_util_extract_config.box_hor_size, 
                    crires_util_extract_config.spec_hsize, 
                    crires_util_extract_config.kappa,
                    crires_util_extract_config.closing_hs,
                    crires_util_extract_config.clean_rate,
                    crires_util_extract_config.rej_left,
                    crires_util_extract_config.rej_right,
                    extr_spec_starty, extr_spec_stopy,
                    i+1,
                    crires_util_extract_config.tot_ndit,
                    crires_util_extract_config.period,
                    &(crires_util_extract_config.qc_specpos[i]), 
                    &(crires_util_extract_config.qc_specwrec[i]), 
                    &(crires_util_extract_config.qc_specwopt[i]), 
                    &(crires_util_extract_config.qc_specoptmed[i]),
                    &(crires_util_extract_config.qc_s2nmed[i]),
                    &(prof[i]),
                    &(bgmap[i])) ;
            cpl_msg_info(__func__, "Chip number %d FWHM Computation", i+1) ;
            if (crires_extract_qc_fwhm(cpl_imagelist_get(imlist, i),
                        prof[i],
                        &(crires_util_extract_config.qc_fwhm_comb_pix[i]),
                        &(crires_util_extract_config.qc_fwhm_comb_as[i]),
                        &(crires_util_extract_config.qc_fwhm_prof_pix[i]),
                        &(crires_util_extract_config.qc_fwhm_prof_as[i]),
                        &(crires_util_extract_config.qc_fwhm_diff[i])) == -1) {
                cpl_msg_warning(__func__, "Failed for FWHM computation") ;
                crires_util_extract_config.qc_fwhm_comb_pix[i] = -1.0 ;
                crires_util_extract_config.qc_fwhm_comb_as[i] = -1.0 ;
                crires_util_extract_config.qc_fwhm_prof_pix[i] = -1.0 ;
                crires_util_extract_config.qc_fwhm_prof_as[i] = -1.0 ;
                crires_util_extract_config.qc_fwhm_diff[i] = -1.0 ;
            }
        } else if (crires_util_extract_config.mode == 2) {
            /* Get the correction for the position */
            if (i+1 == 1) 
                ly = crires_get_detector_ly1(crires_util_extract_config.period);
            if (i+1 == 2) 
                ly = crires_get_detector_ly2(crires_util_extract_config.period);
            if (i+1 == 3) 
                ly = crires_get_detector_ly3(crires_util_extract_config.period);
            if (i+1 == 4) 
                ly = crires_get_detector_ly4(crires_util_extract_config.period);
            /* If ly not found, use the bottom of the image */
            if (ly < 0) ly = 1 ;

            /* Where is the Y extraction zone ? */
            y_pos = "" ;
            if (i+1 == 1) y_pos = crires_util_extract_config.y_pos_c1 ;
            if (i+1 == 2) y_pos = crires_util_extract_config.y_pos_c2 ;
            if (i+1 == 3) y_pos = crires_util_extract_config.y_pos_c3 ;
            if (i+1 == 4) y_pos = crires_util_extract_config.y_pos_c4 ;

            /* Try to parse the user specified positions */
            if (!strcmp(y_pos, "")) {
                ypos_vec = NULL ;
            } else {
                if ((ypos_vec = crires_parse_y_positions(y_pos)) == NULL) {
                    cpl_msg_warning(__func__,
                            "Cannot parse the y_positions value : %s", y_pos) ;
                    ypos_vec = NULL ;
                }
            } 
            
            if (ypos_vec == NULL) {
                cpl_msg_info(__func__, "Collapse the full image along Y") ;
                starty = 1 ;
                stopy = cpl_image_get_size_y(cpl_imagelist_get(imlist, i)) ;
            } else {
                cpl_msg_info(__func__, "Collapse the area around Y=%g",
                        cpl_vector_get(ypos_vec, 0)) ;
                /* TODO : Only the first passed value is used */
                starty = (int)cpl_vector_get(ypos_vec, 0) - 
                    (int)(crires_util_extract_config.y_width/2.0) ;
                stopy = (int)cpl_vector_get(ypos_vec, 0) +
                    (int)(crires_util_extract_config.y_width/2.0) ;
                cpl_vector_delete(ypos_vec) ;
                starty -= ly-1 ;
                stopy -= ly-1 ;
            }

            /* Extract lines */
            ext_vec = crires_extract_lines(cpl_imagelist_get(imlist, i),
                    starty, stopy,
                    crires_util_extract_config.rej_left,
                    crires_util_extract_config.rej_right) ;
            if (ext_vec != NULL) {
                ext[i] = crires_extract_gen_tab(ext_vec, NULL, NULL, NULL, 
                        NULL, NULL, NULL) ;     
                crires_util_extract_config.qc_specpos[i] =
                    (starty+stopy)/2 + (ly-1) ;
                crires_util_extract_config.qc_specwrec[i] = starty-stopy+1;
                cpl_vector_delete(ext_vec) ;
            }
        } else {
            cpl_msg_warning(__func__, "Unsuported mode: %d", 
                    crires_util_extract_config.mode) ;
        }

        /* Plot the result if required */
        if ((ext[i] != NULL) && (crires_util_extract_config.display == i+1)) {
            ext_vec = cpl_vector_wrap(cpl_table_get_nrow(ext[i]),
            cpl_table_get_data_double(ext[i], CRIRES_COL_EXTRACT_INT_RECT));
            cpl_plot_vector(
"set grid;set xlabel 'Position (pixels)';set ylabel 'Intensity RECT (ADU/sec)';",
            "t 'Extracted Spectrum RECT' w lines", "", ext_vec) ;
            cpl_vector_unwrap(ext_vec) ;
            ext_vec = cpl_vector_wrap(cpl_table_get_nrow(ext[i]),
            cpl_table_get_data_double(ext[i], CRIRES_COL_EXTRACT_INT_OPT)) ;
            cpl_plot_vector(
"set grid;set xlabel 'Position (pixels)';set ylabel 'Intensity OPT (ADU/sec)';",
            "t 'Extracted Spectrum OPT' w lines", "", ext_vec) ;
            cpl_vector_unwrap(ext_vec) ;
        }
        cpl_msg_indent_less() ;
    }
    cpl_imagelist_delete(imlist) ;
    if (contrib != NULL) cpl_imagelist_delete(contrib) ;
    cpl_msg_indent_less() ;

    /* Reconstruct chips in windowing mode using detector 2 */
    if (crires_util_extract_config.win_mode == 1) {
        if (ext[1] != NULL) {
            ext[0] = cpl_table_duplicate(ext[1]) ;
            cpl_table_set_size(ext[0], 0) ;
            ext[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(ext[0]) ;
        }
        if (prof[1] != NULL) {
            prof[0] = cpl_image_duplicate(prof[1]) ;
            cpl_image_multiply_scalar(prof[0], 0.0) ;
            prof[CRIRES_NB_DETECTORS-1] = cpl_image_duplicate(prof[0]) ;
        }
        if (bgmap[1] != NULL) {
            bgmap[0] = cpl_image_duplicate(bgmap[1]) ;
            cpl_image_multiply_scalar(bgmap[0], 0.0) ;
            bgmap[CRIRES_NB_DETECTORS-1] = cpl_image_duplicate(bgmap[0]) ;
        }
    }

    /* Test that the spectrum is at the same place in all detectors */
    for (i=1 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Skip some detectors in windowing mode */
        if ((i==1 || i==CRIRES_NB_DETECTORS-1) && 
                (crires_util_extract_config.win_mode == 1)) {
            continue ;
        }
        if (crires_util_extract_config.qc_specpos[i-1] > 0 && 
                crires_util_extract_config.qc_specpos[i] > 0 &&
                fabs(crires_util_extract_config.qc_specpos[i-1] - 
                    crires_util_extract_config.qc_specpos[i]) > 
                CRIRES_SPEC_POS_TOLERANCE) {
            cpl_msg_warning(__func__, 
    "The spectrum positions in chip %d and chip %d are too different: %d -> %d",
                    i, i+1, crires_util_extract_config.qc_specpos[i-1], 
                    crires_util_extract_config.qc_specpos[i]) ;
        }
    }
 
    /* Create the profile */
    profile = cpl_imagelist_new() ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if (prof[i] != NULL) cpl_imagelist_set(profile, prof[i], i) ;
        else {
            cpl_imagelist_delete(profile) ;
            profile = NULL ;
            for (j=i+1 ; j<CRIRES_NB_DETECTORS ; j++) 
                if (prof[j] != NULL) cpl_image_delete(prof[j]) ;
            break ;
        }
    }
   
    /* Create the bg maps */
    bgmaps = cpl_imagelist_new() ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if (bgmap[i] != NULL) cpl_imagelist_set(bgmaps, bgmap[i], i) ;
        else {
            cpl_imagelist_delete(bgmaps) ;
            bgmaps = NULL ;
            for (j=i+1 ; j<CRIRES_NB_DETECTORS ; j++) 
                if (bgmap[j] != NULL) cpl_image_delete(bgmap[j]) ;
            break ;
        }
    }
        
    /* Save the result */
    cpl_msg_info(__func__, "Save the products") ;
    cpl_msg_indent_more() ;
    if (crires_util_extract_save((const cpl_table **)ext, profile, bgmaps, 
                parlist, frameset) == -1) {
        cpl_msg_error(__func__, "Cannot save products") ;
        cpl_msg_indent_less() ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            if (ext[i] != NULL) cpl_table_delete(ext[i]) ;
        }
        if (profile != NULL) cpl_imagelist_delete(profile) ;
        if (bgmaps != NULL) cpl_imagelist_delete(bgmaps) ;
        return -1 ;
    }
    cpl_msg_indent_less() ;
   
    /* Deallocate */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if (ext[i] != NULL) cpl_table_delete(ext[i]) ;
    }
    if (profile != NULL) cpl_imagelist_delete(profile) ;
    if (bgmaps != NULL) cpl_imagelist_delete(bgmaps) ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    ext         the tables with the extracted spectrum
  @param    profile     the image list with the profiles or NULL
  @param    bgmaps      the image list with the bg maps or NULL
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_extract_save(
        const cpl_table             **  ext,
        const cpl_imagelist         *   profile,
        const cpl_imagelist         *   bgmaps,
        const cpl_parameterlist     *   parlist,
        cpl_frameset                *   set)
{
    cpl_propertylist    **  qclists ;
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   inputlist ;
    const char          *   recipe_name = "crires_util_extract" ;
    int                     i ;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECPOS",
                crires_util_extract_config.qc_specpos[i]) ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECWREC",
                crires_util_extract_config.qc_specwrec[i]) ;
        cpl_propertylist_append_int(qclists[i], "ESO QC SPECWOPT",
                crires_util_extract_config.qc_specwopt[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC SIGNAL MED",
                crires_util_extract_config.qc_specoptmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC S2NMED",
                crires_util_extract_config.qc_s2nmed[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMPIX COMBINED",
                crires_util_extract_config.qc_fwhm_comb_pix[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMARC COMBINED",
                crires_util_extract_config.qc_fwhm_comb_as[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMPIX PROFILE",
                crires_util_extract_config.qc_fwhm_prof_pix[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHMARC PROFILE",
                crires_util_extract_config.qc_fwhm_prof_as[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHM DIFF",
                crires_util_extract_config.qc_fwhm_diff[i]) ;

        /* Propagate some keywords from input raw frame extensions */
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), i+1,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist, 
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    if (profile != NULL) {
        /* Write the image */
        crires_image_save(set,
                parlist,
                set,
                profile,
                recipe_name,
                CRIRES_OBS_EXTRACT_PROFILE_IMA,
                CRIRES_PROTYPE_PROFILE,
                crires_util_extract_config.period,
                NULL,
                (const cpl_propertylist**)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_extract_profile.fits") ;
    }

    if (bgmaps!= NULL) {
        /* Write the image */
        crires_image_save(set,
                parlist,
                set,
                bgmaps,
                recipe_name,
                CRIRES_OBS_EXTRACT_BGMAP_IMA,
                CRIRES_PROTYPE_BGD_MAP,
                crires_util_extract_config.period,
                NULL,
                (const cpl_propertylist**)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_extract_bgmap.fits") ;
    }

    /* Write the table */
    crires_table_save(set,
            parlist,
            set,
            ext,
            recipe_name,
            CRIRES_EXTRACT_PIX_TAB,
            CRIRES_PROTYPE_SPEC_PIX,
            NULL,
            (const cpl_propertylist**)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_extract_extracted.fits") ;

    /* Free and return */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    return  0;
}

